#!/usr/bin/env python3

"""
Show chapter x of indexed gzipped moby dick


Usage:

    python read_moby_dick.py <Chapter>
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

from GSGR import GSGR
import sys
import time

my_Reader = GSGR("./Moby_Dick_indexed.gz")

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(__doc__)
    else:
        try:
            chap_num = int(sys.argv[1])
        except:
            chap_num = sys.argv[1]
        print(
            """
    Reading indexed gzip and retrieving chapter {0}
            """.format(
                chap_num
            )
        )
        s = time.time()
        print(
            """{0}

    Took {1:.5f} seconds to retrieve chapter

            """.format(
                my_Reader.read_block(chap_num), time.time() - s
            )
        )



import sys
import os

sys.path.append(os.path.abspath("."))
import pymzml.run as run
from pymzml.spec import Spectrum, Chromatogram
import random
import statistics as stat
import unittest
import test_file_paths
from pprint import pprint

try:
    import numpy as np

    numpy_installed = True
except:
    numpy_installed = False


class SpectrumTest(unittest.TestCase):
    """ """

    def assertPeaksIdentical(
        self, peaks1, peaks2, mult=1, msg=None, measured_precision=5e-6
    ):
        self.assertEqual(
            len(peaks1), len(peaks2), msg="List have different number of peaks!"
        )
        for x in range(len(peaks1)):
            self.assertAlmostEqual(
                peaks1[x][0], peaks2[x][0], msg=msg, delta=measured_precision
            )
            self.assertAlmostEqual(
                peaks1[x][1] * mult, peaks2[x][1], msg=msg, delta=5e-4
            )

    def setUp(self):
        """ """
        self.paths = test_file_paths.paths
        path = self.paths[2]
        self.Run = run.Reader(path)
        self.spec = self.Run[6]

    def test_init(self):
        """ """
        self.assertEqual(self.spec.measured_precision, 5e-6)
        self.assertEqual(self.spec.ns, "{http://psi.hupo.org/ms/mzml}")
        self.assertIsNotNone(self.spec.element)

    def test_mz(self):
        """ """
        mz = self.spec.mz
        self.spec.mz = [1, 2, 3]
        self.assertCountEqual(self.spec.mz, [1, 2, 3])
        Run64 = run.Reader(self.paths[0])
        spec64 = Run64[2]
        mz64 = spec64.mz

    def test_i(self):
        """ """
        i = self.spec.i
        self.spec.i = [1, 2, 3]
        self.assertCountEqual(self.spec.i, [1, 2, 3])

    def test_decode_to_numpy(self):
        """ """
        if numpy_installed:
            test_array = [100.00087, 100.00238, 100.00389, 100.00541, 119.27174]
            b64_array = "cgDIQjgByEL+AchCxQLIQiGL7kIjjO5CJY3uQieO7kIpj+5CK5DuQi2R7kIvku5CMpPuQjeU7kI5le5CO5buQj2X7kK6QgJDTUMCQ+FDAkN0RAJDB0UCQ5pFAkMuRgJDwUYCQ1RHAkPoRwJDe0gCQw9JAkOiSQJDj/YGQyr3BkPF9wZDYfgGQ/z4BkOX+QZDM/oGQ876BkNp+wZDBfwGQ6L8BkM9/QZD2P0GQ3T+BkPmBRVDmgYVQ04HFUMCCBVDtggVQ2oJFUMfChVD0woVQ4cLFUM7DBVD8AwVQ6QNFUNYDhVDDA8VQ8EPFUN1EBVDKREVQ/UFFkOrBhZDYQcWQxcIFkPNCBZDgwkWQzkKFkPvChZDpQsWQ1sMFkMRDRZDxw0WQ34OFkM0DxZDw6gZQ4CpGUM9qhlD+aoZQ7arGUNzrBlDL60ZQ+ytGUOprhlDZq8ZQyKwGUPfsBlDnLEZQ0iiGkMGoxpDxaMaQ4OkGkNCpRpDAKYaQ7+mGkN9pxpDPKgaQ/+oGkO9qRpDfKoaQzqrGkO0o0dDy6RHQ+OlR0P6pkdDEahHQympR0NAqkdDWKtHQ2+sR0OHrUdDnq5HQ7avR0PNsEdDaT1RQ5U+UUPAP1FD7EBRQxhCUUNEQ1FDcERRQ5xFUUPHRlFD9EdRQyBJUUNLSlFDd0tRQ3p0bEPidWxDS3dsQ7N4bEMbemxDg3tsQ+t8bENTfmxDvH9sQySBbEONgmxD9YNsQ12FbEPGhmxDDn6EQ+R+hEO5f4RDj4CEQ2WBhEM6goRDEIOEQ+aDhEO7hIRDkYWEQ2eGhEM8h4RDEoiEQ5x6hUN0e4VDTHyFQyR9hUP9fYVD1X6FQ61/hUOFgIVDXYGFQzWChUMNg4VD5YOFQ72EhUOVhYVDlnaGQ3B3hkNLeIZDJXmGQwB6hkPaeoZDtHuGQ498hkNpfYZDRH6GQx5/hkP5f4ZD04CGQ66BhkOIgoZDY4OGQz2EhkMM9oZD6PaGQ8T3hkOf+IZDe/mGQ1f6hkMy+4ZDDvyGQ+r8hkPF/YZDof6GQ33/hkNYAIdDNAGHQxACh0PrAodDG9CIQ/vQiEPb0YhDu9KIQ5vTiEN71IhDXNWIQzzWiEMc14hD/NeIQ9zYiEO82YhDnNqIQ33biENvQYxDV0KMQ0BDjEMpRIxDEUWMQ/pFjEPjRoxDy0eMQ7RIjEOdSYxDhkqMQ25LjENXTIxDT3+MQzmAjEMigYxDC4KMQ/WCjEPeg4xDx4SMQ7CFjEOahoxDg4eMQ2yIjENWiYxDP4qMQyiLjEMSjIxD+4yMQ+WNjEMhAI1DCwGNQ/YBjUPgAo1DywONQ7UEjUOgBY1DigaNQ3UHjUNgCI1DSgmNQzUKjUMfC41DCgyNQ/UMjUP0fY1D4H6NQ8x/jUO3gI1Do4GNQ4+CjUN7g41DZ4SNQ1KFjUM+ho1DKoeNQxaIjUMCiY1D7YmNQ9mKjUO6eY5DqXqOQ5d7jkOFfI5Dc32OQ2F+jkNQf45DPoCOQyyBjkMago5DCYOOQ/eDjkPlhI5D1IWOQ8KGjkOwh45Dn4iOQ3n6jkNo+45DWPyOQ0f9jkM3/o5DJv+OQxYAj0MFAY9D9QGPQ+QCj0PUA49DwwSPQ7IFj0OiBo9DkQePQ4EIj0P7f5VD+4CVQ/uBlUP7gpVD+4OVQ/uElUP7hZVD+4aVQ/uHlUP7iJVD/ImVQ/yKlUP8i5VD/IyVQ/yNlUP8jpVD/I+VQ4AtmkOMLppDmC+aQ6UwmkOxMZpDvTKaQ8kzmkPVNJpD4jWaQ+42mkP6N5pDBjmaQxM6mkMfO5pDbF3DQ+pew0NpYMND52HDQw=="

            out_array = self.spec._decode_to_numpy(
                b64_array, 4292, "32-bit float", "no compression"
            )
            val1 = round(float(test_array[1]), 3)
            val2 = round(float(out_array[1]), 3)
            self.assertEqual(val1, val2)

    def test_decode_to_tuple(self):
        """ """
        test_array = "123"
        b64_array = "cgDIQjgByEL+AchCxQLIQiGL7kIjjO5CJY3uQieO7kIpj+5CK5DuQi2R7kIvku5CMpPuQjeU7kI5le5CO5buQj2X7kK6QgJDTUMCQ+FDAkN0RAJDB0UCQ5pFAkMuRgJDwUYCQ1RHAkPoRwJDe0gCQw9JAkOiSQJDj/YGQyr3BkPF9wZDYfgGQ/z4BkOX+QZDM/oGQ876BkNp+wZDBfwGQ6L8BkM9/QZD2P0GQ3T+BkPmBRVDmgYVQ04HFUMCCBVDtggVQ2oJFUMfChVD0woVQ4cLFUM7DBVD8AwVQ6QNFUNYDhVDDA8VQ8EPFUN1EBVDKREVQ/UFFkOrBhZDYQcWQxcIFkPNCBZDgwkWQzkKFkPvChZDpQsWQ1sMFkMRDRZDxw0WQ34OFkM0DxZDw6gZQ4CpGUM9qhlD+aoZQ7arGUNzrBlDL60ZQ+ytGUOprhlDZq8ZQyKwGUPfsBlDnLEZQ0iiGkMGoxpDxaMaQ4OkGkNCpRpDAKYaQ7+mGkN9pxpDPKgaQ/+oGkO9qRpDfKoaQzqrGkO0o0dDy6RHQ+OlR0P6pkdDEahHQympR0NAqkdDWKtHQ2+sR0OHrUdDnq5HQ7avR0PNsEdDaT1RQ5U+UUPAP1FD7EBRQxhCUUNEQ1FDcERRQ5xFUUPHRlFD9EdRQyBJUUNLSlFDd0tRQ3p0bEPidWxDS3dsQ7N4bEMbemxDg3tsQ+t8bENTfmxDvH9sQySBbEONgmxD9YNsQ12FbEPGhmxDDn6EQ+R+hEO5f4RDj4CEQ2WBhEM6goRDEIOEQ+aDhEO7hIRDkYWEQ2eGhEM8h4RDEoiEQ5x6hUN0e4VDTHyFQyR9hUP9fYVD1X6FQ61/hUOFgIVDXYGFQzWChUMNg4VD5YOFQ72EhUOVhYVDlnaGQ3B3hkNLeIZDJXmGQwB6hkPaeoZDtHuGQ498hkNpfYZDRH6GQx5/hkP5f4ZD04CGQ66BhkOIgoZDY4OGQz2EhkMM9oZD6PaGQ8T3hkOf+IZDe/mGQ1f6hkMy+4ZDDvyGQ+r8hkPF/YZDof6GQ33/hkNYAIdDNAGHQxACh0PrAodDG9CIQ/vQiEPb0YhDu9KIQ5vTiEN71IhDXNWIQzzWiEMc14hD/NeIQ9zYiEO82YhDnNqIQ33biENvQYxDV0KMQ0BDjEMpRIxDEUWMQ/pFjEPjRoxDy0eMQ7RIjEOdSYxDhkqMQ25LjENXTIxDT3+MQzmAjEMigYxDC4KMQ/WCjEPeg4xDx4SMQ7CFjEOahoxDg4eMQ2yIjENWiYxDP4qMQyiLjEMSjIxD+4yMQ+WNjEMhAI1DCwGNQ/YBjUPgAo1DywONQ7UEjUOgBY1DigaNQ3UHjUNgCI1DSgmNQzUKjUMfC41DCgyNQ/UMjUP0fY1D4H6NQ8x/jUO3gI1Do4GNQ4+CjUN7g41DZ4SNQ1KFjUM+ho1DKoeNQxaIjUMCiY1D7YmNQ9mKjUO6eY5DqXqOQ5d7jkOFfI5Dc32OQ2F+jkNQf45DPoCOQyyBjkMago5DCYOOQ/eDjkPlhI5D1IWOQ8KGjkOwh45Dn4iOQ3n6jkNo+45DWPyOQ0f9jkM3/o5DJv+OQxYAj0MFAY9D9QGPQ+QCj0PUA49DwwSPQ7IFj0OiBo9DkQePQ4EIj0P7f5VD+4CVQ/uBlUP7gpVD+4OVQ/uElUP7hZVD+4aVQ/uHlUP7iJVD/ImVQ/yKlUP8i5VD/IyVQ/yNlUP8jpVD/I+VQ4AtmkOMLppDmC+aQ6UwmkOxMZpDvTKaQ8kzmkPVNJpD4jWaQ+42mkP6N5pDBjmaQxM6mkMfO5pDbF3DQ+pew0NpYMND52HDQw=="
        out_array = self.spec._decode_to_tuple(
            b64_array, 343, "32-bit float", "no compression"
        )
        # self.assertEqual(out_array, test_array)

    def test_reprofile_peaks(self):
        """ """
        r_peaks = self.spec.peaks("reprofiled")
        self.assertIsNotNone(r_peaks)
        # self.assertEqual(r_peaks, [(1, 10),(2, 20),(3, 30),(4, 40)])

    def test_centroid_peaks(self):
        """ """
        self.spec.set_peaks([(1, 10), (2, 20), (3, 30), (4, 40)], "centroided")
        c_peaks = self.spec.peaks("centroided")
        self.assertPeaksIdentical(c_peaks, [(1, 10), (2, 20), (3, 30), (4, 40)])

    def test_centroid_peaks_for_real(self):
        """ """
        new_peaks = [
            (99.999, 0.1),
            (99.9995, 1),
            (100.0000, 10),
            (100.0005, 1),
            (100.001, 0.1),
        ]
        self.spec.set_peaks(new_peaks, "raw")
        self.spec.measured_precision = 1e-5
        peaks = self.spec.peaks("centroided")
        self.assertPeaksIdentical(peaks, new_peaks)

    def test_median(self):
        array = [1, 4, 6, 7, 2, 5, 7, 23.324, 5.0, -4.4, 0]
        res = self.spec._median(array)
        res_correct = stat.median(array)
        self.assertEqual(res_correct, res)

    def test_add(self):
        """ """
        spec = self.Run[6]
        p1 = spec.peaks("reprofiled")
        spec += spec
        p2 = spec.peaks("reprofiled")
        self.assertPeaksIdentical(
            p1, p2, mult=2
        )  # , msg='List 1 : {0}\nList 2:{1}'.format(p1, p2))

    def test_sub(self):
        spec = self.Run[6]
        p1 = spec.peaks("reprofiled")
        spec -= spec
        p2 = spec.peaks("reprofiled")

        self.assertPeaksIdentical(
            p1, p2, mult=0
        )  # , msg='List 1 : {0}\nList 2:{1}'.format(p1, p2))

    def test_mult(self):
        """ """
        new_peaks = [(1, 10), (2, 20)]
        self.spec.set_peaks(new_peaks, "raw")
        self.spec * 2
        self.assertCountEqual(self.spec.peaks("raw")[0], (1, 20))
        self.assertCountEqual(self.spec.peaks("raw")[1], (2, 40))

    def test_div(self):
        """ """
        new_peaks = [(1, 10), (2, 20)]
        self.spec.set_peaks(new_peaks, "raw")
        self.spec / 2
        self.assertCountEqual(self.spec.peaks("raw")[0], (1, 5))
        self.assertCountEqual(self.spec.peaks("raw")[1], (2, 10))

    def test_add_specs_to_empty_spec(self):
        spec1 = Spectrum()
        spec2 = Spectrum()
        spec2.set_peaks([(100, 200)], "raw")
        spec1 += spec2
        centroided_mz = spec1.peaks("centroided")[:, 0]
        centroided_i = spec1.peaks("centroided")[:, 1]
        assert np.allclose(centroided_mz, [100], rtol=5e-6)
        assert np.allclose(centroided_i, [200], atol=0.002)

    def test_add_tow_custom_specs(self):
        spec1 = Spectrum()
        spec2 = Spectrum()
        spec1.set_peaks([(100, 200)], "raw")
        spec2.set_peaks([(100, 200), (200, 300)], "raw")
        spec1 += spec2
        centroided_mz = spec1.peaks("centroided")[:, 0]
        centroided_i = spec1.peaks("centroided")[:, 1]
        assert np.allclose(centroided_mz, [100, 200], rtol=5e-6)
        assert np.allclose(centroided_i, [400, 300], atol=0.002)

    def test_average_spectra(self):
        spec0 = Spectrum()
        spec1 = Spectrum()
        spec2 = Spectrum()

        spec1.set_peaks(np.array([(100, 200.0)]), "centroided")
        spec2.set_peaks(np.array([(100, 200.0), (200, 300.0)]), "centroided")

        scaled1 = spec1 / spec1.peaks("centroided")[:, 1].sum()
        spec0 += scaled1

        scaled2 = spec2 / spec2.peaks("centroided")[:, 1].sum()
        spec0 += scaled2

        centroided_peaks = spec0.peaks("centroided")
        assert np.allclose(centroided_peaks[:, 0], [100, 200])
        assert np.allclose(centroided_peaks[:, 1], [1.4, 0.6], atol=0.0001)

    def test_reduce(self):
        """ """
        new_peaks = [(50, 100), (102, 300), (400, 20), (401, 207)]
        # self.spec.peaks = new_peaks
        self.spec.set_peaks(new_peaks, "raw")
        peaks = self.spec.peaks("raw")
        self.spec.reduce(mz_range=(100, 400))
        self.assertPeaksIdentical(self.spec.peaks("raw"), [(102, 300), (400, 20)])

    def test_highest_peaks(self):
        """ """
        h_peaks = self.spec.highest_peaks(3)
        high_i = [i for mz, i in h_peaks]
        self.assertEqual(
            sorted([i for mz, i in self.spec.peaks("centroided")])[-3:], high_i
        )

    def test_extreme_values(self):
        """ """
        new_peaks = [(1, 10000), (10000, 1), (500, 600), (600, 500)]
        self.spec.set_peaks(new_peaks, "raw")
        extreme_vals_i = self.spec.extreme_values("i")
        extreme_vals_mz = self.spec.extreme_values("mz")
        self.assertEqual(extreme_vals_mz, (1, 10000))
        self.assertEqual(extreme_vals_i, (1, 10000))

    def test_has_peak(self):
        """ """
        i = self.spec.peaks("centroided")[1][0]
        hits = self.spec.has_peak(i)
        hits2 = self.spec.has_peak(i + 5e-6)
        self.assertAlmostEqual(hits[0][0], i)
        self.assertIsNotNone(hits2[0][0], i)

    def test_transform_mz(self):
        """ """
        mz = 213.33333333333333
        tMZ = self.spec.transform_mz(mz)
        self.assertEqual(tMZ, int(round(mz * self.spec.internal_precision)))

    def test_has_overlapping_peak(self):
        """ """
        overlaps = []
        mzs = self.spec.mz
        for mz in mzs:
            overlaps.append(self.spec.has_overlapping_peak(mz))
        self.assertTrue(any(overlaps))

    def test_similarity_to(self):
        """ """
        cos = self.spec.similarity_to(self.spec)
        self.assertEqual(round(cos), 1)

        # cos in this part to high, is other cos right though???
        spec2 = self.spec
        spec2.set_peaks([(1, 2), (3, 4)], "raw")
        cos = self.spec.similarity_to(spec2)
        self.assertLess(cos, 1)

    def test_peaks_are_set(self):
        spec = self.spec
        spec.set_peaks([(1000, 10)], "raw")
        self.assertEqual(spec.mz, [1000])
        self.assertEqual(spec.i, [10])
        self.assertPeaksIdentical(spec.peaks("raw"), [(1000, 10)])

    def test_spec_has_peak_within_precision(self):
        """ """
        pass

    def test_getitem_index_gzip(self):
        idx_gzip = self.paths[2]
        run_idx_gzip = run.Reader(idx_gzip)
        randInt = 6
        spec = run_idx_gzip[randInt]
        self.assertIsInstance(spec, Spectrum)
        self.assertEqual(spec.ID, randInt)
        self.assertIsInstance(run_idx_gzip["TIC"], Chromatogram)

    def test_getitem_standard_gzip_1(self):
        gzip = self.paths[1]
        run_gzip = run.Reader(gzip)
        self.assertIsInstance(run_gzip[1], Spectrum)

    def test_getitem_standard_gzip_tic(self):
        gzip = self.paths[1]
        run_gzip = run.Reader(gzip)
        self.assertIsInstance(run_gzip["TIC"], Chromatogram)

    def test_getitem_standard_mzml_random(self):
        mzml = self.paths[0]
        RunMzml = run.Reader(mzml)
        randInt = 4
        self.assertIsInstance(RunMzml[randInt], Spectrum)
        self.assertIsInstance(RunMzml["TIC"], Chromatogram)

    def test_getitem_2_times_standard_mzml(self):
        mzml = self.paths[0]
        RunMzml = run.Reader(mzml)
        randInt = 10
        spec1 = RunMzml[randInt]
        self.assertIn(spec1.ID, RunMzml.info["offset_dict"])
        spec2 = RunMzml[randInt]
        self.assertIsInstance(RunMzml[randInt], Spectrum)
        self.assertEqual(spec1.ID, spec2.ID)

    def test_getitems_end_standard_mzml(self):
        mzml = self.paths[0]
        RunMzml = run.Reader(mzml)
        spec = RunMzml[11]
        self.assertEqual(spec.ID, 11)

    def test_getitem_index_mzml(self):
        idx_mzml = self.paths[1]
        run_idx_mzml = run.Reader(idx_mzml)
        randInt = 1
        spec = run_idx_mzml[randInt]
        self.assertIsInstance(spec, Spectrum)
        self.assertEqual(spec.ID, randInt)
        self.assertIsInstance(run_idx_mzml["TIC"], Chromatogram)

    def test_getitem_index_mzml_last(self):
        idx_mzml = self.paths[0]
        run_idx_mzml = run.Reader(idx_mzml)
        spec = run_idx_mzml[11]
        self.assertIsInstance(spec, Spectrum)
        self.assertEqual(spec.ID, 11)

    def test_getitem_index_mzml_double(self):
        idx_mzml = self.paths[0]
        run_idx_mzml = run.Reader(idx_mzml)
        self.assertIsInstance(run_idx_mzml[4], Spectrum)
        self.assertIn(4, run_idx_mzml.info["offset_dict"])
        self.assertIsInstance(run_idx_mzml[4], Spectrum)

    def test_get_search_spec_with_string(self):
        mzml = self.paths[0]
        Reader = run.Reader(mzml)
        string = "scan=2"
        self.assertIsInstance(Reader[string], Spectrum)

    def test_get_element_by_name(self):
        search_name = "scan window lower limit"
        ele = self.spec.get_element_by_name(search_name)
        self.assertEqual(ele.get("name"), search_name)

    def test_get_element_by_path(self):
        elements = self.spec.get_element_by_path(hooks=["scanList", "cvParam"])
        self.assertIsInstance(elements, list)
        for ele in elements:
            self.assertIn("cvParam", ele.tag)
        elements = self.spec.get_element_by_path(
            ["scanList", "scan", "scanWindowList", "scanWindow", "cvParam"]
        )
        self.assertIsInstance(elements, list)
        for ele in elements:
            self.assertIn("cvParam", ele.tag)

    def test_peaks(self):
        peaks = self.spec.peaks("raw")
        self.assertPeaksIdentical(
            peaks[:5],
            [
                [7.00487061e01, 6.13473730e03],
                [7.00657654e01, 4.08554297e04],
                [7.10610352e01, 2.26755859e04],
                [7.20450134e01, 6.51679375e04],
                [7.20813980e01, 3.49662469e05],
            ],
        )

    def test_estimate_noise_mean_spec_6(self):
        spec = self.Run[6]
        self.assertAlmostEqual(90407.70511570283, spec.estimated_noise_level("mean"))

    def test_remove_noise_mean_spec_6(self):
        spec = self.Run[6]
        spec.remove_noise("mean")
        self.assertPeaksIdentical(
            spec.peaks("centroided")[:5],
            [
                [7.20813980e01, 3.49662469e05],
                [7.40606766e01, 1.18710039e05],
                [7.40970306e01, 1.35170670e07],
                [7.51003647e01, 6.28674688e05],
                [8.40448761e01, 1.36323156e05],
            ],
        )

    def test_remove_noise_median_spec_6(self):
        spec = self.Run[6]
        spec.remove_noise("median")
        self.assertPeaksIdentical(
            spec.peaks("centroided")[:5],
            [
                [7.00657654e01, 4.08554297e04],
                [7.10610352e01, 2.26755859e04],
                [7.20450134e01, 6.51679375e04],
                [7.20813980e01, 3.49662469e05],
                [7.30847321e01, 2.11254648e04],
            ],
        )

    def test_remove_noise_mad_spec_6(self):
        spec = self.Run[6]
        p1 = spec.peaks("centroided")
        spec.remove_noise("mad")
        p2 = spec.peaks("centroided")
        self.assertPeaksIdentical(
            p2[:5],
            [
                [7.00657654e01, 4.08554297e04],
                [7.10610352e01, 2.26755859e04],
                [7.20450134e01, 6.51679375e04],
                [7.20813980e01, 3.49662469e05],
                [7.30847321e01, 2.11254648e04],
            ],
        )

    def test_remove_noise_s2n_noise_level_spec_6(self):
        spec = self.Run[6]
        p1 = spec.peaks("centroided")
        spec.remove_noise(noise_level=1e5, signal_to_noise_threshold=2.0)
        p2 = spec.peaks("centroided")
        self.assertPeaksIdentical(
            p2[:5],
            [
                [7.20813980e01, 3.49662469e05],
                [7.40970306e01, 1.35170670e07],
                [7.51003647e01, 6.28674688e05],
                [8.60605240e01, 3.99662406e05],
                [8.60969086e01, 5.50612938e05],
            ],
        )

    def test_remove_noise_s2n_median_spec_6(self):
        spec = self.Run[6]
        p1 = spec.peaks("centroided")
        spec.remove_noise(mode="mad", signal_to_noise_threshold=5.0)
        p2 = spec.peaks("centroided")
        self.assertPeaksIdentical(
            p2[:5],
            [
                [7.00657654e01, 4.08554297e04],
                [7.20450134e01, 6.51679375e04],
                [7.20813980e01, 3.49662469e05],
                [7.40606766e01, 1.18710039e05],
                [7.40970306e01, 1.35170670e07],
            ],
        )

    def test_transformed_mz_with_error(self):
        tmz = self.spec.transformed_mz_with_error
        self.assertIsInstance(tmz, dict)
        self.assertIsNotNone(len(tmz))

    def test_transformed_peaks(self):
        spec = self.Run[6]
        tpeaks = spec.transformed_peaks
        self.assertIsInstance(tpeaks, list)
        self.assertIsNotNone(len(tpeaks))
        known_tPeaks = [
            (700487, 6134.7373046875),
            (700658, 40855.4296875),
            (710610, 22675.5859375),
            (720450, 65167.9375),
            (720814, 349662.46875),
        ]
        self.assertPeaksIdentical(tpeaks[:5], known_tPeaks)

    def test_t_mz_set(self):
        spec = self.Run[6]
        tmzs = spec.t_mz_set
        self.assertIsInstance(tmzs, set)
        self.assertEqual(len(tmzs), 23081)

    # def test_precursosr(self):
    #     ms2_spec = self.Run[3]
    #     prec =ms2_spec.precursors
    #     self.assertIsNotNone(len(prec))
    #     self.assertIsInstance(prec, list)
    #     self.assertEqual(prec, ['199'])

    def test_TIC(self):
        tic = self.spec.TIC
        self.assertIsNotNone(tic)
        self.assertIsInstance(tic, float)
        self.assertEqual(tic, 9.6721256e07)

    def test_scan_time(self):
        scan_time, unit = self.spec.scan_time
        self.assertIsNotNone(scan_time)
        self.assertIsInstance(scan_time, float)
        self.assertEqual(scan_time, 0.023756566)

    def test_get_all_arrays_in_spec(self):
        assert self.spec.get_all_arrays_in_spec() == ["m/z array", "intensity array"]

    def test_get_array(self):
        # import pdb;pdb.set_trace()
        assert (self.spec.mz == self.spec.get_array("m/z array")).all()

    def test_get_tims_tof_ion_mobility(self):
        assert self.spec.get_tims_tof_ion_mobility() is None


if __name__ == "__main__":
    unittest.main(verbosity=3)



#!/usr/bin/env python3
# encoding: utf-8

'''

Originally created for Ursgal ( https://github.com/ursgal/ursgal )


'''

import glob
import os

if __name__ == '__main__':
    print('''
        Formatting example scripts into rst files for the docs
''')
    # input()
    example_script_path = os.path.join(
        os.path.dirname(__file__),
        os.pardir,
        'example_scripts',
        '*.py',
    )
    print(example_script_path)
    for example_script in glob.glob(example_script_path):
        if os.path.exists(example_script) is False:
            continue
        basename = os.path.basename(example_script)
        print(
            'Reading: {0}'.format(example_script)
        )
        file_path = os.path.join(
            os.path.dirname(__file__),
            'source',
            'code_inc',
            '{0}',
        )
        with open( file_path.format( basename.replace('.py', '.inc')), 'w') as o:
            print('''.. code-block:: python\n''', file=o)
            with open( example_script ) as infile:
                for line in infile:
                    print('\t{0}'.format( line.rstrip()), file=o)





#!/usr/bin/env python
import sys
import pymzml
from collections import defaultdict as ddict


def main(mzml_file):
    """
    Basic example script to demonstrate the usage of pymzML. Requires a mzML
    file as first argument.

    usage:
        ./simple_parser_v2.py <path_to_mzml_file>

    Note:

        This script uses the old syntax where the MS level can be queried as a
        key (Spectrum['ms level']). The current syntax can be found in
        simple_parser.py

    """
    run = pymzml.run.Reader(mzml_file)
    # print( run[10000].keys() )
    stats = ddict(int)
    for n, spec in enumerate(run):
        print(
            "Spectrum {0}, MS level {ms_level}".format(n, ms_level=spec["ms level"]),
            end="\r",
        )
        # the old method to obtain peaks from the Spectrum class
        stats[spec.ID] = len(spec.centroidedPeaks)

    print("Parsed {0} spectra from file {1}".format(len(stats.keys()), mzml_file))
    print()


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(main.__doc__)
        exit()
    mzml_file = sys.argv[1]
    main(mzml_file)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Interface for binary streams of uncompressed mzML.

@author: Sylvain Le Bon
"""
from io import TextIOWrapper

from .. import regex_patterns
from .standardMzml import StandardMzml


class BytesMzml(StandardMzml):
    def __init__(self, binary, encoding, build_index_from_scratch=False):
        """
        Initalize Wrapper object for standard mzML files.

        Arguments:
            path (str)     : path to the file
            encoding (str) : encoding of the file
        """
        self.binary = binary
        self.file_handler = self.get_file_handler(encoding)
        self.offset_dict = dict()
        self.spec_open = regex_patterns.SPECTRUM_OPEN_PATTERN
        self.spec_close = regex_patterns.SPECTRUM_CLOSE_PATTERN
        if build_index_from_scratch is True:
            seeker = self.get_binary_file_handler()
            self._build_index_from_scratch(seeker)
            seeker.close()

    def get_binary_file_handler(self):
        self.binary.seek(0)
        return self.binary

    def get_file_handler(self, encoding):
        return TextIOWrapper(self.binary, encoding=encoding)



#!/usr/bin/env python3

import sys
import os
from pymzml.utils.utils import index_gzip
from pymzml.run import Reader


def main(mzml_path, out_path):
    """
    Create and indexed gzip mzML file from a plain mzML.

    Usage: python3 gzip_mzml.py <path/to/mzml> <path/to/output>
    """
    with open(mzml_path) as fin:
        fin.seek(0, 2)
        max_offset_len = fin.tell()
        max_spec_no = Reader(mzml_path).get_spectrum_count() + 10

    index_gzip(
        mzml_path, out_path, max_idx=max_spec_no, idx_len=len(str(max_offset_len))
    )


if __name__ == "__main__":
    if len(sys.argv) > 2:
        main(sys.argv[1], sys.argv[2])
    else:
        print(main.__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Part of pymzml test cases
"""
import os
from pymzml.file_interface import FileInterface
from pymzml.file_classes.standardGzip import StandardGzip
from pymzml.file_classes.indexedGzip import IndexedGzip
from pymzml.file_classes.standardMzml import StandardMzml
import unittest
import test_file_paths


class FileInterfaceTest(unittest.TestCase):
    """ """

    def setUp(self):
        """ """
        self.paths = test_file_paths.paths
        test_file = self.paths[1]
        self.File = FileInterface(test_file, "latin-1")

    def tearDown(self):
        """ """
        self.File.close()

    def test_init(self):
        """ """
        self.assertIsNotNone(self.File.file_handler)

    def test_open(self):
        """ """
        self.assertIsInstance(self.File.file_handler, StandardGzip)
        self.File.close()
        self.File = FileInterface(self.paths[0], "latin-1")
        self.assertIsInstance(self.File.file_handler, StandardMzml)
        self.File.close()
        self.File = FileInterface(self.paths[2], "latin-1")
        self.assertIsInstance(self.File.file_handler, IndexedGzip)
        self.File.close()


if __name__ == "__main__":
    unittest.main(verbosity=3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Part of pymzml test cases
"""
import os
from pymzml.utils.GSGW import GSGW
import unittest
import zlib
import struct


class GSGWTest(unittest.TestCase):
    """
    TODO: Write messages for assertions
    """

    def setUp(self):
        self.paths = [os.path.join(os.path.dirname(__file__), "data", "unittest.mzml")]
        self.Writer = GSGW(
            self.paths[0],
            max_idx=80,
            max_idx_len=8,
            max_offset_len=8,
            output_path=os.path.abspath(
                os.path.join(".", "tests", "data", "unittest.mzml")
            ),
        )

    def tearDown(self):
        """ """
        self.Writer.close()
        os.remove(os.path.abspath(os.path.join(".", "tests", "data", "unittest.mzml")))

    def test_init(self):
        self.assertEqual(self.Writer.crc32, 0)
        self.assertEqual(self.Writer.crc32, 0)
        self.assertEqual(
            self.Writer.file_name,
            os.path.abspath(
                os.path.join(
                    # os.path.dirname(__file__),
                    # 'Test',
                    "tests",
                    "data",
                    "unittest.mzml",
                )
            ),
        )

    def test_write_gen_header_index(self):
        self.Writer._write_gen_header(Index=True)
        self.Writer.close()
        file = open(self.paths[0], "rb")
        header = file.read()
        file.close()

        # Normal stuff
        self.assertEqual(header[:2], b"\x1f\x8b")  # Gzip magic bytes
        self.assertEqual(header[2], 8)  # comp
        self.assertEqual(header[3], 16)  # comment flags set
        # self.assertEqual(header[5:9], b'') # time, how to check?
        self.assertEqual(header[8], 2)  # xfl
        self.assertEqual(header[9], 3)  # os

        # extra field with index
        self.assertEqual(header[10:12], b"FU")  # magic index bytes
        self.assertEqual(header[12], 1)  # version

    def test_write_gen_header_no_index(self):
        self.Writer._write_gen_header(Index=False)
        self.Writer.close()
        file = open(self.paths[0], "rb")
        header = file.read()
        file.close()
        # Normal stuff
        self.assertEqual(header[:2], b"\x1f\x8b")
        self.assertEqual(header[2], 8)  # comp
        self.assertEqual(header[3], 0)  # comment flags set
        # self.assertEqual(header[5:9], b'') # time, how to check?
        self.assertEqual(header[8], 2)  # xfl
        self.assertEqual(header[9], 3)  # os
        self.assertEqual(len(header), 10)

    def test_allocate_index_bytes(self):
        self.Writer._allocate_index_bytes()
        self.Writer.close()
        file = open(self.paths[0], "rb")
        data = file.read()
        print(data, data[1], len(data))
        file.close()
        max_idx_num = 80
        max_idx = 8
        max_offset = 8
        self.assertEqual(len(data), ((max_idx + max_offset) * max_idx_num) + 1)

        self.assertEqual(len(data), max_idx_num * (max_idx + max_offset) + 1)

    def test_write_data(self):
        test_string = b"AAAAAAAAbbbbbbbbCCCCCCCC"
        self.Writer._write_data(test_string)
        self.Writer.close()
        Decomp = zlib.decompressobj(-zlib.MAX_WBITS)
        file = open(
            os.path.join(os.path.dirname(__file__), "data", "unittest.mzml"), "rb"
        )
        data = file.read()
        file.close()
        compData = data[:-8]
        crc = struct.unpack("<L", data[-8:-4])[0]
        isize = struct.unpack("<L", data[-4:])[0]
        self.assertEqual(Decomp.decompress(compData), test_string)
        self.assertEqual(crc, zlib.crc32(test_string))
        self.assertEqual(isize, len(test_string) % 2**32)

    def test_add_data(self):
        test_string = b"AAAAAAAAbbbbbbbbCCCCCCCCdddddddd"
        self.Writer.add_data(test_string, 1)
        self.Writer.close()
        file = open(self.paths[0], "rb")
        data = file.read()
        file.close()
        index = self.Writer.index
        self.assertIn(1, index)
        test_file = os.path.join(
            os.path.dirname(__file__),
            "data",
            "uncompressed",
            "CF_07062012_pH8_2_3A.mzML",
        )
        self.Writer = GSGW(test_file, max_idx=80, max_idx_len=8, max_offset_len=8)
        self.Writer.add_data(test_string, "a")
        self.Writer.close()
        file = open(self.paths[0], "rb")
        data = file.read()
        file.close()
        index = self.Writer.index
        self.assertIn("a", index)

    def test_write_index(self):
        test_string = b"AAAAAAAAbbbbbbbbCCCCCCCCdddddddd"

        # with int as id
        # TODO assertions
        self.Writer.add_data(test_string, 1)
        self.Writer.write_index()
        self.Writer.close()
        file = open(self.paths[0], "rb")
        data = file.read()
        file.close()
        index = data[
            15 : 15 + (80 * 16) + 1
        ]  # dont read header (first 15 bytes) and read idx_num * (idx_len + offset_len bytes) and zero termination
        identifier = index[:8].decode("latin-1")
        offset = index[8:16].decode("latin-1")
        self.assertEqual(identifier.strip("\xAC"), "1")
        self.assertIsInstance(offset.strip("\xAC"), str)


if __name__ == "__main__":
    unittest.main(verbosity=3)



#!usr/bin/env python3
# -*- coding: latin-1 -*-
"""
The spectrum class offers a python object for mass spectrometry data.
The spectrum object holds the basic information of the spectrum and offers
methods to interrogate properties of the spectrum.
Data, i.e. mass over charge (m/z) and intensity decoding is performed on demand
and can be accessed via their properties, e.g. :py:attr:`~pymzml.spec.Spectrum.peaks`.

The Spectrum class is used in the :py:class:`~pymzml.run.Reader` class.
There each spectrum is accessible as a spectrum object.

Theoretical spectra can also be created using the setter functions.
For example, m/z values, intensities, and peaks can be set by the
corresponding properties: :py:attr:`pymzml.spec.Spectrum.mz`,
:py:attr:`pymzml.spec.Spectrum.i`, :py:attr:`pymzml.spec.Spectrum.peaks`.

Similar to the spectrum class, the chromatogram class allows interrogation
with profile data (time, intensity) in an total ion chromatogram.
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

import math
import re
import sys
import warnings
import xml.etree.ElementTree as ElementTree
import zlib
from base64 import b64decode as b64dec
from collections import defaultdict as ddict
from functools import lru_cache
from operator import itemgetter as itemgetter
from struct import unpack

import numpy as np

try:
    DECON_DEP = True
    from ms_deisotope.deconvolution import deconvolute_peaks
    from ms_peak_picker import simple_peak
except (ImportError, ModuleNotFoundError) as e:
    DECON_DEP = False
    pass
from . import regex_patterns
from .decoder import MSDecoder
from .obo import OboTranslator

PROTON = 1.00727646677
ISOTOPE_AVERAGE_DIFFERENCE = 1.002


class MS_Spectrum(object):
    """
    General spectrum class for data handling.
    """

    def _read_accessions(self):
        """Set all required variables for this spectrum."""
        self.accessions = {}
        for element in self.element.iter():
            accession = element.get("accession")
            name = element.get("name")
            if accession is not None:
                self.accessions[name] = accession
        if "profile spectrum" in self.accessions.keys():
            self._profile = True

    def get_element_by_name(self, name):
        """
        Get element from the original tree by it's unit name.

        Arguments:
            name (str): unit name of the mzml element.

        Keyword Arguments:
            obo_version (str, optional): obo version number.

        """
        iterator = self.element.iter()
        return_ele = None
        for ele in iterator:
            if ele.get("name", default=None) == name:
                return_ele = ele
                break
        return return_ele

    def get_element_by_path(self, hooks):
        """
        Find elements in spectrum by its path.

        Arguments:
            hooks (list): list of parent elements for the target element.

        Returns:
            elements (list): list of XML objects
            found in the path

        Example:

            To access cvParam in scanWindow tag:

            >>> spec.get_element_by_path(['scanList', 'scan', 'scanWindowList',
            ...     'scanWindow', 'cvParam'])

        """
        return_ele = None
        if len(hooks) > 0:
            path_array = ["."]
            for hook in hooks:
                path_array.append("{ns}{hook}".format(ns=self.ns, hook=hook))
            path = "/".join(path_array)
            return_ele = self.element.findall(path)

        return return_ele

    def _register(self, decoded_tuple):
        d_type, array = decoded_tuple
        if d_type == "mz":
            self._mz = array
        elif d_type == "i":
            self._i = array
        elif d_type == "time":
            self._time = array
        else:
            raise Exception("Unknown data Type ({0})".format(d_type))

    def _get_encoding_parameters(self, array_type):
        """
        Find the correct parameter for decoding and return them as tuple.

        Arguments:
            array_type (str): data type of the array, e.g. m/z, time or
                intensity
        Returns:
            data (str)           : encoded data
            comp (str)           : compression method
            d_type (str)         : data type
            d_array_length (str) : length of the data array
        """
        numpress_encoding = False

        b_data_string = "./{ns}binaryDataArrayList/{ns}binaryDataArray/{ns}cvParam[@name='{name}']/..".format(
            ns=self.ns, name=array_type
        )
        float_type_string = "./{ns}cvParam[@accession='{Acc}']"

        b_data_array = self.element.find(b_data_string)
        if not b_data_array:
            # non-standard data array
            b_data_string = "./{ns}binaryDataArrayList/{ns}binaryDataArray/{ns}cvParam[@value='{value}']/..".format(
                ns=self.ns, value=array_type
            )
            b_data_array = self.element.find(b_data_string)

        comp = []
        if b_data_array:
            for cvParam in b_data_array.iterfind("./{ns}cvParam".format(ns=self.ns)):
                if "compression" in cvParam.get("name"):
                    if "numpress" in cvParam.get("name").lower():
                        numpress_encoding = True
                    comp.append(cvParam.get("name"))
                d_array_length = self.element.get("defaultArrayLength")
            if not numpress_encoding:
                try:
                    # 32-bit float
                    d_type = b_data_array.find(
                        float_type_string.format(
                            ns=self.ns,
                            Acc=self.obo_translator["32-bit float"]["id"],
                        )
                    ).get("name")
                except:
                    try:
                        # 64-bit Float
                        d_type = b_data_array.find(
                            float_type_string.format(
                                ns=self.ns,
                                Acc=self.obo_translator["64-bit float"]["id"],
                            )
                        ).get("name")
                    except:
                        try:
                            # 32-bit integer
                            d_type = b_data_array.find(
                                float_type_string.format(
                                    ns=self.ns,
                                    Acc=self.obo_translator["32-bit integer"]["id"],
                                )
                            ).get("name")
                        except:
                            try:
                                # 64-bit integer
                                d_type = b_data_array.find(
                                    float_type_string.format(
                                        ns=self.ns,
                                        Acc=self.obo_translator["64-bit integer"]["id"],
                                    )
                                ).get("name")
                            except:
                                # null-terminated ASCII string
                                d_type = b_data_array.find(
                                    float_type_string.format(
                                        ns=self.ns,
                                        Acc=self.obo_translator[
                                            "null-terminated ASCII string"
                                        ]["id"],
                                    )
                                ).get("name")
            else:
                # compression is numpress, dont need data type here
                d_type = None
            data = b_data_array.find("./{ns}binary".format(ns=self.ns))
            if data is not None:
                data = data.text
        else:
            data = None
            d_array_length = 0
            d_type = "64-bit float"
        if data is not None:
            data = data.encode("utf-8")
        else:
            data = ""
        return (data, d_array_length, d_type, comp)

    @property
    def measured_precision(self):
        """
        Set the measured and internal precision.

        Returns:
            value (float): measured Precision (e.g. 5e-6)
        """
        return self._measured_precision

    @measured_precision.setter
    def measured_precision(self, value):
        self._measured_precision = value
        self.internal_precision = int(round(50000.0 / (value * 1e6)))
        return

    def _decode_to_numpy(self, data, d_array_length, data_type, comp):
        """
        Decode the b64 encoded and packed strings from data as numpy arrays.

        Returns:
            data (np.ndarray): Returns the unpacked data as a tuple. Returns an
                               empty list if there is no raw data or raises an
                               exception if data could not be decoded.

        d_array_length just for compatibility
        """
        out_data = b64dec(data)
        if len(out_data) != 0:
            if "zlib" in comp or "zlib compression" in comp:
                out_data = zlib.decompress(out_data)
            if (
                "ms-np-linear" in comp
                or "ms-np-pic" in comp
                or "ms-np-slof" in comp
                or "MS-Numpress linear prediction compression" in comp
                or "MS-Numpress short logged float compression" in comp
            ):
                out_data = self._decodeNumpress_to_array(out_data, comp)
            if data_type == "32-bit float":
                # one character code may be sufficient too (f)
                f_type = np.float32
                out_data = np.frombuffer(out_data, f_type)
            elif data_type == "64-bit float":
                # one character code may be sufficient too (d)
                f_type = np.float64
                out_data = np.frombuffer(out_data, f_type)
            elif data_type == "32-bit integer":
                # one character code may be sufficient too (i)
                i_type = np.int32
                out_data = np.frombuffer(out_data, i_type)
            elif data_type == "64-bit integer":
                # one character code may be sufficient too (l)
                i_type = np.int64
                out_data = np.frombuffer(out_data, i_type)
            # TODO elif data_type == "null-terminated ASCII string":
            else:
                raise ValueError(f"Unsupported data type: {data_type}")
        else:
            out_data = np.array([])
        return out_data

    def _decode_to_tuple(self, data, d_array_length, float_type, comp):
        """
        Decode b64 encoded and packed strings.

        Returns:
            data (tuple): Returns the unpacked data as a tuple.
                Returns an empty list if there is no raw data or
                raises an exception if data could not be decoded.
        """
        dec_data = b64dec(data)
        if len(dec_data) != 0:
            if "zlib" in comp or "zlib compression" in comp:
                dec_data = zlib.decompress(dec_data)
            if set(["ms-np-linear", "ms-np-pic", "ms-np-slof"]) & set(comp):
                self._decodeNumpress(data, comp)
            # else:
            #     print(
            #         'New data compression ({0}) detected, cant decompress'.format(
            #             comp
            #         )
            #     )
            #     sys.exit(1)
            if float_type == "32-bit float":
                f_type = "f"
            elif float_type == "64-bit float":
                f_type = "d"
            fmt = "{endian}{array_length}{float_type}".format(
                endian="<", array_length=d_array_length, float_type=f_type
            )
            ret_data = unpack(fmt, dec_data)
        else:
            ret_data = []
        return ret_data

    def _decodeNumpress_to_array(self, data, compression):
        """
        Decode golomb-rice encoded data (aka numpress encoded data).

        Arguments:
            data (str)        : Encoded data string
            compression (str) : Decompression algorithm to be used
                (valid are 'ms-np-linear', 'ms-np-pic', 'ms-np-slof')

        Returns:
            array (list): Returns the unpacked data as an array of floats.

        """
        result = []
        comp_ms_tags = [self.calling_instance.OT[comp]["id"] for comp in compression]
        data = np.frombuffer(data, dtype=np.uint8)
        if "MS:1002312" in comp_ms_tags:
            result = MSDecoder.decode_linear(data)
        elif "MS:1002313" in comp_ms_tags:
            result = MSDecoder.decode_pic(data)
        elif "MS:1002314" in comp_ms_tags:
            result = MSDecoder.decode_slof(data)
        return result

    def _median(self, data):
        """
        Compute median.

        Arguments:
            data (list): list of numeric values

        Returns:
            median (float): median of the input data
        """
        return np.median(data)

    def to_string(self, encoding="latin-1", method="xml"):
        """
        Return string representation of the xml element the
        spectrum was initialized with.

        Keyword Arguments:
            encoding (str) : text encoding of the returned string.\n
                             Default is latin-1.
            method (str)   : text format of the returned string.\n
                             Default is xml, alternatives are html and text.

        Returns:
            element (str)  : xml string representation of the spectrum.
        """
        return ElementTree.tostring(self.element, encoding=encoding, method=method)


class Spectrum(MS_Spectrum):
    """
    Spectrum class which inherits from class :py:attr:`pymzml.spec.MS_Spectrum`

    Arguments:
        element (xml.etree.ElementTree.Element): spectrum as xml element

    Keyword Arguments:
        measured_precision (float): in ppm, i.e. 5e-6 equals to 5 ppm.

    """

    def __init__(
        self,
        element=ElementTree.Element(""),
        measured_precision=5e-6,
        *,
        obo_version=None,
    ):
        __slots__ = [
            "_centroided_peaks",
            "_centroided_peaks_sorted_by_i",
            "_deconvoluted_peaks",
            "_extreme_values",
            "_i",
            "_ID",
            "_id_dict",
            "_index",
            "_measured_precision",
            "_peaks",
            "_profile",
            "_reprofiled_peaks",
            "_t_mass_set",
            "_t_mz_set",
            "_time",
            "_transformed_mass_with_error",
            "_transformed_mz_with_error",
            "_transformed_peaks" "calling_instance" "element",
            "internal_precision" "noise_level_estimate",
            "selected_precursors",
        ]

        self._centroided_peaks = None
        self._centroided_peaks_sorted_by_i = None
        self._extreme_values = None
        self._i = None
        self._ID = None
        self._id_dict = None
        self._index = None
        self._ms_level = None
        self._mz = None
        self._peak_dict = {
            "raw": None,
            "centroided": None,
            "reprofiled": None,
            "deconvoluted": None,
        }
        self._selected_precursors = None
        self._profile = None
        self.reprofiled = False
        self._reprofiled_peaks = None
        self._scan_time = None
        self._scan_time_unit = None
        self._scan_time_in_minutes = None
        self._t_mass_set = None
        self._t_mz_set = None
        self._TIC = None
        self._precursors = None
        self._transformed_mass_with_error = None
        self._transformed_mz_with_error = None
        self._transformed_peaks = None
        self.obo_translator = OboTranslator.from_cache(obo_version)
        self.element = element
        self.measured_precision = measured_precision
        self.noise_level_estimate = {}

        self.ns = ""
        if self.element:
            self.ns = (
                re.match(r"\{.*\}", element.tag).group(0)
                if re.match(r"\{.*\}", element.tag)
                else ""
            )

        self._decode = self._decode_to_numpy
        self._array = np.array
        self._ms_deisotop_warning_printed = False

    def __del__(self):
        """
        Clear self.element to limit RAM usage
        """
        if self.element:
            self.element.clear()

    def __add__(self, other_spec):
        """
        Adds two pymzml spectra

        Arguments:
            other_spec (Spectrum): spectrum to add to the current spectrum

        Returns:
            self (Spectrum): reference to the edited spectrum

        Example:

        >>> import pymzml
        >>> s = pymzml.spec.Spectrum( measuredPrescision = 20e-6 )
        >>> file_to_read = "../mzML_example_files/xy.mzML.gz"
        >>> run = pymzml.run.Reader(
        ...     file_to_read ,
        ...     MS1_Precision = 5e-6 ,
        ...     MSn_Precision = 20e-6
        ... )
        >>> for spec in run:
        ...     s += spec

        """
        assert isinstance(other_spec, Spectrum)
        if self._peak_dict["reprofiled"] is None:
            reprofiled = self._reprofile_Peaks()
            self.set_peaks(reprofiled, "reprofiled")
        if other_spec._peak_dict["reprofiled"] is None:
            other_spec.set_peaks(other_spec._reprofile_Peaks(), "reprofiled")
        for mz, i in other_spec.peaks("reprofiled"):
            self._peak_dict["reprofiled"][mz] += i
        return self

    def __sub__(self, other_spec):
        """
        Subtracts two pymzml spectra.

        Arguments:
            other_spec (spec.Spectrum): spectrum to subtract from the current
                spectrum

        Returns:
            self (spec.Spectrum): returns self after other_spec was subtracted
        """
        assert isinstance(other_spec, Spectrum)
        if self._peak_dict["reprofiled"] is None:
            self.set_peaks(self._reprofile_Peaks(), "reprofiled")
        if other_spec._peak_dict["reprofiled"] is None:
            other_spec.set_peaks(other_spec._reprofile_Peaks(), "reprofiled")
        for mz, i in other_spec.peaks("reprofiled"):
            self._peak_dict["reprofiled"][mz] -= i
        self.set_peaks(None, "centroided")
        self.set_peaks(None, "raw")
        return self

    def __mul__(self, value):
        """
        Multiplies each intensity with a float, i.e. scales the spectrum.

        Arguments:
            value (int, float): value to multiply the intensities with

        Returns:
            self (spec.Spectrum): returns self after intensities were scaled
                by value
        """
        assert isinstance(value, (int, float))
        if self._peak_dict["raw"] is not None:
            self.set_peaks(
                np.column_stack(
                    (self.peaks("raw")[:, 0], self.peaks("raw")[:, 1] * value)
                ),
                "raw",
            )
        if self._peak_dict["centroided"] is not None:
            self.set_peaks(
                np.column_stack(
                    (self.centroided_peaks[:, 0], self.centroided_peaks[:, 1] * value)
                ),
                "centroided",
            )
        if self._peak_dict["reprofiled"] is not None:
            for mz in self._peak_dict["reprofiled"].keys():
                self._peak_dict["reprofiled"][mz] *= float(value)
        return self

    def __truediv__(self, value):
        """
        Divides each intensity by a float, i.e. scales the spectrum.

        Arguments:
            value (int, float): value to divide the intensities by

        Returns:
            self (spec.Spectrum): returns self after intensities were scaled
                by value.
        """
        if self._peak_dict["reprofiled"] is not None:
            for mz, i in self.peaks("reprofiled"):
                self._peak_dict["reprofiled"][mz] /= float(value)
        if self._peak_dict["raw"] is not None:
            if len(self._peak_dict["raw"]) != 0:
                self.set_peaks(
                    np.column_stack(
                        (
                            self.peaks("raw")[:, 0],
                            self.peaks("raw")[:, 1] / float(value),
                        )
                    ),
                    "raw",
                )
        if self._peak_dict["centroided"] is not None:
            peaks = self._peak_dict["centroided"]
            scaled_peaks = peaks[:, 1] / value
            peaks[:, 1] = scaled_peaks
            self._peak_dict["centroided"] = peaks
        return self

    def __div__(self, value):
        """
        Integer division is the same as __truediv__ for this class
        """
        return self.__truediv__(value)

    def __repr__(self):
        """
        Returns representative string for a spectrum object class
        """
        return "<__main__.Spectrum object with native ID {0} at {1}>".format(
            self.ID, hex(id(self))
        )

    def __str__(self):
        """
        Returns representative string for a spectrum object class
        """
        return "<__main__.Spectrum object with native ID {0} at {1}>".format(
            self.ID, hex(id(self))
        )

    @lru_cache()
    def __getitem__(self, accession):
        """
        Access spectrum XML information by tag name

        Args:
            accession(str): name of the XML tag

        Returns:
            value (float or str): value of the XML tag
        """
        #  TODO implement cache???
        if accession == "id":
            return_val = self.ID
        else:
            if not accession.startswith("MS:"):
                try:
                    accession = self.obo_translator[accession]["id"]
                except TypeError:
                    accession = "---"
            search_string = './/*[@accession="{0}"]'.format(accession)
            elements = []
            for x in self.element.iterfind(search_string):
                val = x.attrib.get("value", "")
                try:
                    val = float(val)
                except:
                    pass
                elements.append(val)

            if len(elements) == 0:
                return_val = None
            elif len(elements) == 1:
                return_val = elements[0]
            else:
                return_val = elements
        if return_val == "":
            return_val = True
        return return_val

    def get(self, acc, default=None):
        """Mimic dicts get function.

        Args:
            acc (str): accession or obo tag to return
            default (None, optional): default value if acc is not found
        """
        val = self[acc]
        if val is None:
            val = default
        return val

    def __contains__(self, value):
        """Check if MS tag or name can be found in spectrum.

        Args:
            value (str): MS tag or OBO name

        Returns:
            bool
        """
        r = False
        if self[value] is not None:
            r = True
        return r

    @property
    def measured_precision(self):
        """
        Sets the measured and internal precision

        Returns:
            value (float): measured precision (e.g. 5e-6)
        """
        return self._measured_precision

    @measured_precision.setter
    def measured_precision(self, value):
        self._measured_precision = value
        self.internal_precision = int(round(50000.0 / (value * 1e6)))
        return

    @property
    def t_mz_set(self):
        """
        Creates a set of integers out of transformed m/z values
        (including all values in the defined imprecision).
        This is used to accelerate has_peak function and similar.

        Returns:
            t_mz_set (set): set of transformed m/z values
        """
        if self._t_mz_set is None:
            self._t_mz_set = set()
            for mz, i in self.peaks("centroided"):
                self._t_mz_set |= set(
                    range(
                        int(
                            round(
                                (mz - (mz * self.measured_precision))
                                * self.internal_precision
                            )
                        ),
                        int(
                            round(
                                (mz + (mz * self.measured_precision))
                                * self.internal_precision
                            )
                        )
                        + 1,
                    )
                )
        return self._t_mz_set

    @property
    def transformed_mz_with_error(self):
        """
        Returns transformed m/z value with error

        Returns:
            tmz values (dict): Transformed m/z values in dictionary\n
                {\n
                m/z_with_error : [(m/z,intensity), ...], ...\n
                }\n
        """
        if self._transformed_mz_with_error is None:
            self._transformed_mz_with_error = ddict(list)
            for mz, i in self.peaks("centroided"):
                for t_mz_with_error in range(
                    int(
                        round(
                            (mz - (mz * self.measured_precision))
                            * self.internal_precision
                        )
                    ),
                    int(
                        round(
                            (mz + (mz * self.measured_precision))
                            * self.internal_precision
                        )
                    )
                    + 1,
                ):
                    self._transformed_mz_with_error[t_mz_with_error].append((mz, i))
        return self._transformed_mz_with_error

    @property
    def transformed_peaks(self):
        """
        m/z value is multiplied by the internal precision.

        Returns:
            Transformed peaks (list): Returns a list of peaks (tuples of mz and
            intensity).
            Float m/z values are adjusted by the internal precision
            to integers.
        """
        if self._transformed_peaks is None:
            self._transformed_peaks = [
                (self.transform_mz(mz), i) for mz, i in self.peaks("centroided")
            ]
        return self._transformed_peaks

    @property
    def TIC(self):
        """
        Property to access the total ion current for this spectrum.

        Returns:
            TIC (float): Total Ion Current of the spectrum.
        """
        if self._TIC is None:
            self._TIC = float(
                self.element.find(
                    "./{ns}cvParam[@accession='MS:1000285']".format(ns=self.ns)
                ).get("value")
            )  # put hardcoded MS tags in minimum.py???
        return self._TIC

    @property
    def ID(self):
        """
        Access the native id (last number in the id attribute) of the spectrum.

        Returns:
            ID (str): native ID of the spectrum
        """
        if self._ID is None:
            if self.element:
                match = regex_patterns.SPECTRUM_ID_PATTERN.search(
                    self.element.get("id", None)
                )
                if match:
                    try:
                        self._ID = int(match.group(1))
                    except ValueError:
                        self._ID = match.group(1)
                else:
                    self._ID = self.element.get("id")
        return self._ID

    @property
    def id_dict(self):
        """
        Access to all entries stored the id attribute of a spectrum.

        Returns:
            id_dict (dict): key value pairs for all entries in id attribute of a spectrum
        """
        if self._id_dict is None:
            tuples = []
            match = regex_patterns.SPECTRUM_PATTERN3.match(self.element.attrib["id"])
            if match is not None:
                captures = match.captures(1)
                for element in captures:
                    k, v = element.strip().split("=")
                    v = int(v)
                    tuples.append([k, v])
                self._id_dict = dict(tuples)
            else:
                self._id_dict = {}
        return self._id_dict

    @property
    def index(self):
        """
        Access the index of the spectrum.

        Returns:
            index (int): index of the spectrum

        Note:
            This does not necessarily correspond to the native spectrum ID
        """
        if self._index is None:
            self._index = self.element.get("index")
            try:
                self._index = int(self._index)
            except:
                pass
        return self._index

    @property
    def ms_level(self):
        """
        Property to access the ms level.

        Returns:
            ms_level (int):
        """
        if self._ms_level is None:
            sub_element = self.element.find(
                ".//{ns}cvParam[@accession='MS:1000511']".format(ns=self.ns)
            )
            if sub_element is not None:
                self._ms_level = int(
                    sub_element.get("value")
                )  # put hardcoded MS tags in minimum.py???
        return self._ms_level

    @property
    def scan_time(self):
        """
        Property to access the retention time and retention time unit.
        Please note, that we do not assume the retention time unit,
        if it is not correctly defined in the mzML.
        It is set to 'unicorns' in this case.

        Returns:
            scan_time (float):
            scan_time_unit (str):
        """
        if self._scan_time is None or self._scan_time_unit is None:
            scan_time_ele = self.element.find(
                ".//*[@accession='MS:1000016']".format(ns=self.ns)
            )
            if scan_time_ele is not None:
                self._scan_time = float(scan_time_ele.attrib.get("value"))
                self._scan_time_unit = scan_time_ele.get("unitName", "unicorns")
        return self._scan_time, self._scan_time_unit

    # @property
    def scan_time_in_minutes(self):
        """
        Property to access the retention time in minutes.
        If the retention time unit is defined within the mzML,
        the retention time is converted into minutes and returned
        without the unit.

        Returns:
            scan_time (float):
        """
        if self._scan_time_in_minutes is None:
            self._scan_time, time_unit = self.scan_time
            if self._scan_time_unit.lower() == "millisecond":
                self._scan_time_in_minutes = self._scan_time / 1000.0 / 60.0
            if self._scan_time_unit.lower() == "second":
                self._scan_time_in_minutes = self._scan_time / 60.0
            elif self._scan_time_unit.lower() == "minute":
                self._scan_time_in_minutes = self._scan_time
            elif self._scan_time_unit.lower() == "hour":
                self._scan_time_in_minutes = self._scan_time * 60.0
            else:
                raise Exception("Time unit '{0}' unknown".format(self._scan_time_unit))
        return self._scan_time_in_minutes

    @property
    def selected_precursors(self):
        """
        Property to access the selected precursors of a MS2 spectrum. Returns
        a list of dicts containing the precursors mz and, if available intensity
        and charge for each precursor.

        Returns:
            selected_precursors (list):
        """
        if self._selected_precursors is None:
            selected_precursor_mzs = self.element.findall(
                ".//*[@accession='MS:1000744']"
            )
            selected_precursor_is = self.element.findall(
                ".//*[@accession='MS:1000042']"
            )
            selected_precursor_cs = self.element.findall(
                ".//*[@accession='MS:1000041']"
            )
            precursors = self.element.findall(
                "./{ns}precursorList/{ns}precursor".format(ns=self.ns)
            )

            mz_values = []
            i_values = []
            charges = []
            ids = []
            for obj in selected_precursor_mzs:
                mz = obj.get("value")
                mz_values.append(float(mz))
            for obj in selected_precursor_is:
                i = obj.get("value")
                i_values.append(float(i))
            for obj in selected_precursor_cs:
                c = obj.get("value")
                charges.append(int(c))
            for prec in precursors:
                spec_ref = prec.get("spectrumRef")
                if spec_ref is not None:
                    ids.append(
                        regex_patterns.SPECTRUM_ID_PATTERN.search(spec_ref).group(1)
                    )
                else:
                    ids.append(None)
                # ids.append()
            self._selected_precursors = []
            for pos, mz in enumerate(mz_values):
                dict_2_save = {"mz": mz}
                for key, list_of_values in [
                    ("i", i_values),
                    ("charge", charges),
                    ("precursor id", ids),
                    ("element", precursors),
                ]:
                    try:
                        dict_2_save[key] = list_of_values[pos]
                    except:
                        continue
                self._selected_precursors.append(dict_2_save)

        return self._selected_precursors

    @property
    def precursors(self):
        """
        List the precursor information of this spectrum, if available.
        Returns:
            precursor(list): list of precursor ids for this spectrum.
        """
        self.deprecation_warning(sys._getframe().f_code.co_name)
        if not self._precursors:
            precursors = self.element.findall(
                "./{ns}precursorList/{ns}precursor".format(ns=self.ns)
            )
            self._precursors = []
            for prec in precursors:
                spec_ref = prec.get("spectrumRef")
                self._precursors.append(
                    regex_patterns.SPECTRUM_ID_PATTERN.search(spec_ref).group(1)
                )
        return self._precursors

    def remove_precursor_peak(self):
        peaks = self.peaks("centroided")
        for precursor in self.selected_precursors:
            mz = precursor["mz"]
            hp = self.has_peak(mz)
            if hp:
                for p in hp:
                    peaks = peaks[(peaks[:, 0] != p[0])]
        self.set_peaks(peaks, "centroided")
        self.set_peaks(peaks, "raw")
        return peaks

    @property
    def mz(self):
        """
        Returns the list of m/z values. If the m/z values are encoded, the
        function :func:`~spec.MS_Spectrum._decode` is used to decode the encoded data.
        The mz property can also be set, e.g. for theoretical data.
        However, it is recommended to use the peaks property to set mz and
        intensity tuples at same time.

        Returns:
            mz (list): list of m/z values of spectrum.
        """
        if self._mz is None:
            params = self._get_encoding_parameters("m/z array")
            self._mz = self._decode(*params)
        return self._mz

    @mz.setter
    def mz(self, mz_list):
        """"""
        mz_list = np.array(mz_list, dtype=np.float64)
        mz_list.sort()
        self._mz = mz_list

    @property
    def i(self):
        """
        Returns the list of the intensity values.
        If the intensity values are encoded, the function :func:`~spec.MS_Spectrum._decode`
        is used to decode the encoded data.\n
        The i property can also be set, e.g. for theoretical data. However,
        it is recommended to use the peaks property to set mz and intensity
        tuples at same time.

        Returns
            i (list): list of intensity values from the analyzed spectrum
        """
        if self._i is None:
            params = self._get_encoding_parameters("intensity array")
            self._i = self._decode(*params)
        return self._i

    @i.setter
    def i(self, intensity_list):
        self._i = intensity_list

    def peaks(self, peak_type):
        """
        Decode and return a list of mz/i tuples.

        Args:
            peak_type(str): currently supported types are:
                raw, centroided and reprofiled

        Returns:
            peaks (list or ndarray): list or numpy array of mz/i tuples or arrays
        """
        if self._peak_dict[peak_type] is None:
            if self._peak_dict["raw"] is None:
                mz_params = self._get_encoding_parameters("m/z array")
                i_params = self._get_encoding_parameters("intensity array")
                mz = self._decode(*mz_params)
                i = self._decode(*i_params)
                arr = np.stack((mz, i), axis=-1)
                self._peak_dict[peak_type] = arr
            if peak_type == "raw":
                pass
            elif peak_type == "centroided":
                self._peak_dict["centroided"] = self._centroid_peaks()
            elif peak_type == "reprofiled":
                self._peak_dict["reprofiled"] = self._reprofile_Peaks()
            elif peak_type == "deconvoluted":
                self._peak_dict["deconvoluted"] = self._deconvolute_peaks()
            else:
                raise KeyError

        if not isinstance(self._peak_dict[peak_type], np.ndarray):
            peaks = self._array(self._peak_dict[peak_type])
        else:
            peaks = self._peak_dict[peak_type]
        if peak_type == "reprofiled":
            peaks = list(self._peak_dict[peak_type].items())
            peaks.sort(key=itemgetter(0))
        return peaks

    @lru_cache()
    def get_array(self, arr_name):
        array_params = self._get_encoding_parameters(arr_name)
        array = self._decode(*array_params)
        if len(array) == 0:
            array = None
            _ = self.get_all_arrays_in_spec(not_found_array=arr_name)
        return array

    def get_tims_tof_ion_mobility(
        self, array_name="mean inverse reduced ion mobility array"
    ):
        arr = self.get_array(array_name)
        if arr is None:
            _ = self.get_all_arrays_in_spec(not_found_array=array_name)
        return arr

    def get_all_arrays_in_spec(self, not_found_array=None):
        b_data_string = "./{ns}binaryDataArrayList/{ns}binaryDataArray/{ns}cvParam[@unitCvRef='MS']".format(
            ns=self.ns
        )
        b_data_arrays = self.element.findall(b_data_string)
        array_names = [arr.attrib["name"] for arr in b_data_arrays]
        formatted_array_names = []
        for name in array_names:
            formatted_array_names.append("\t- {name}".format(name=name))
        if not_found_array is not None:
            print(
                "Requested array ({not_found_array}) not found.\nAvailable arrays are:".format(
                    not_found_array=not_found_array
                )
            )
            print("\n".join(formatted_array_names))
        return array_names

    def _deconvolute_peaks(self, *args, **kwargs):
        if DECON_DEP is True:
            peaks = self.peaks("centroided")
            # pack peak matrix into expected structure
            peaks = [simple_peak(p[0], p[1], 0.01) for p in peaks]
            decon_result = deconvolute_peaks(peaks, *args, **kwargs)
            dpeaks = decon_result.peak_set
            # pack deconvoluted peak list into matrix structure
            dpeaks_mat = np.zeros((len(dpeaks), 3), dtype=float)
            for i, dp in enumerate(dpeaks):
                dpeaks_mat[i, :] = dp.neutral_mass, dp.intensity, dp.charge
            return dpeaks_mat
        else:
            if self._ms_deisotop_warning_printed is False:
                print(
                    "ms_deisotope is missing, please install using pip install ms_deisotope"
                )
                self._ms_deisotop_warning_printed = True

    def set_peaks(self, peaks, peak_type):
        """
        Assign a custom peak array of type peak_type

        Args:
            peaks(list or ndarray): list or array of mz/i values
            peak_type(str): Either raw, centroided or reprofiled

        """
        peak_type = peak_type.lower()
        # reset after changing peaks
        self._transformed_mass_with_error = None
        self._transformed_mz_with_error = None
        if peak_type == "raw":
            # if not isinstance(peaks, np.ndarray):
            #     peaks = np.array(peaks)
            self._peak_dict["raw"] = peaks
            try:
                self._mz = self.peaks("raw")[:, 0]
                self._i = self.peaks("raw")[:, 1]
            except IndexError:
                self._mz = np.array([])
                self._i = np.array([])

        elif peak_type == "centroided":
            # if not isinstance(peaks, np.ndarray):
            #     peaks = np.array(peaks)
            self._peak_dict["centroided"] = peaks
            try:
                self._mz = self.peaks("raw")[:, 0]
                self._i = self.peaks("raw")[:, 1]
            except IndexError:
                self._mz = np.array([])
                self._i = np.array([])

        elif peak_type == "reprofiled":
            try:
                self._peak_dict["reprofiled"] = peaks
            except TypeError:
                self._peak_dict["reprofiled"] = None
        elif peak_type == "deconvoluted":
            self._peak_dict["deconvoluted"] = peaks
            try:
                self._mz = self.peaks("raw")[:, 0]
                self._i = self.peaks("raw")[:, 1]
            except IndexError:
                self._mz = np.array([])
        else:
            raise Exception(
                "Peak type is not suppported\n"
                'Choose either "raw", "centroided" or "reprofiled"'
            )

    def _centroid_peaks(self):
        """
        Perform a Gauss fit to centroid the peaks for the property
        centroided_peaks.

        Returns:
            centroided_peaks (list): list of centroided m/z, i tuples
        """

        try:
            profile_ot = self.obo_translator.name.get("profile spectrum", None)
            if profile_ot is None:
                profile_ot = self.obo_translator.name.get("profile mass spectrum", None)
            acc = profile_ot["id"]
            is_profile = (
                True
                if self.element.find(
                    ".//*[@accession='{acc}']".format(ns=self.ns, acc=acc)
                )
                is not None
                else None
            )

        except (TypeError, AttributeError) as e:
            # user creadted spectrum objects without xml and calling instance cant determine if they are reprofiled or not
            is_profile = None

        if is_profile is not None or self.reprofiled:  # check if spec is a profile spec
            tmp = []
            if self._peak_dict["reprofiled"] is not None:
                i_array = [i for mz, i in self.peaks("reprofiled")]
                mz_array = [mz for mz, i in self.peaks("reprofiled")]
            else:
                i_array = self.i
                mz_array = self.mz
            for pos, i in enumerate(i_array[:-1]):
                if pos <= 1:
                    continue
                if 0 < i_array[pos - 1] < i > i_array[pos + 1] > 0:
                    x1 = float(mz_array[pos - 1])
                    y1 = float(i_array[pos - 1])
                    x2 = float(mz_array[pos])
                    y2 = float(i_array[pos])
                    x3 = float(mz_array[pos + 1])
                    y3 = float(i_array[pos + 1])
                    if x2 - x1 > (x3 - x2) * 10 or (x2 - x1) * 10 < x3 - x2:
                        continue
                    if y3 == y1:
                        y3 += 0.01 * y1

                    try:
                        double_log = math.log(y2 / y1) / math.log(y3 / y1)
                        mue = (double_log * (x1 * x1 - x3 * x3) - x1 * x1 + x2 * x2) / (
                            2 * (x2 - x1) - 2 * double_log * (x3 - x1)
                        )
                        c_squarred = (
                            x2 * x2 - x1 * x1 - 2 * x2 * mue + 2 * x1 * mue
                        ) / (2 * math.log(y1 / y2))
                        A = y1 * math.exp((x1 - mue) * (x1 - mue) / (2 * c_squarred))
                    except ZeroDivisionError:
                        continue
                    tmp.append((mue, A))
            return tmp
        else:
            return self.peaks("raw")

    def _reprofile_Peaks(self):
        """
        Performs reprofiling for property reprofiled_peaks.

        Returns:
            reprofiled_peaks (list): list of reprofiled m/z, i tuples
        """
        tmp = ddict(int)
        for mz, i in self.peaks("centroided"):
            # Let the measured precision be 2 sigma of the signal width
            # When using normal distribution
            # FWHM = 2 sqt(2 * ln(2)) sigma = 2.3548 sigma
            s = mz * self.measured_precision * 2  # in before 2
            s2 = s * s
            floor = mz - 5.0 * s  # Gauss curve +- 3 sigma
            ceil = mz + 5.0 * s
            ip = self.internal_precision / 4
            # more spacing, i.e. less points describing the gauss curve
            # -> faster adding
            for _ in range(int(round(floor * ip)), int(round(ceil * ip)) + 1):
                if _ % int(5) == 0:
                    a = float(_) / float(ip)
                    y = i * math.exp(-1 * ((mz - a) * (mz - a)) / (2 * s2))
                    tmp[a] += y
        self.reprofiled = True
        self.set_peaks(None, "centroided")
        return tmp

    def _mz_2_mass(self, mz, charge):
        """
        Calculate the uncharged mass for a given mz value

        Arguments:
            mz (float)  : m/z value
            charge(int) : charge

        Returns:
            mass (float): Returns mass of a given m/z value
        """
        return (mz - PROTON) * charge

    def _set_params_from_reference_group(self, ref_element):
        ref = self.element.find("{ns}referenceableParamGroupRef".format(ns=self.ns))
        if ref is not None:
            ref = ref.get("ref")
        ele = ref_element.find(".//*[@id='{ref}']".format(ref=ref, ns=self.ns))
        if ele is not None and ref == ele.get("id"):
            for param in ele.iter():
                self.element.append(ele)
                acc = param.get("accession")

    # Public functions

    def reduce(self, peak_type="raw", mz_range=(None, None)):
        """
        Remove all m/z values outside the given range.

        Arguments:
            mz_range (tuple): tuple of min, max values

        Returns:
            peaks (list): list of mz, i tuples in the given range.
        """
        arr = self.peaks(peak_type)
        peaks = arr[
            np.where(np.logical_and(arr[:, 0] >= mz_range[0], arr[:, 0] <= mz_range[1]))
        ]
        self.set_peaks(peaks, peak_type)
        return peaks

    def remove_noise(
        self, mode="median", noise_level=None, signal_to_noise_threshold=1.0
    ):
        """
        Function to remove noise from peaks, centroided peaks and reprofiled
        peaks.

        Keyword arguments:
                mode (str): define mode for removing noise. Default = "median"
                (other modes: "mean", "mad")
            noise_level (float): noise threshold
            signal_to_noise_threshold (float): S/N threshold for a peak to be accepted

        Returns:
            reprofiled peaks (list): Returns a list with tuples of
            m/z-intensity pairs above the noise threshold

        """
        # Thanks to JD Hogan for pointing it out!
        callPeaks = self.peaks("raw")
        # callcentPeaks = self.peaks("centroided")
        if noise_level is None:
            noise_level = self.estimated_noise_level(mode=mode)
        if len(self.peaks("centroided")) != 0:
            self._peak_dict["centroided"] = self.peaks("centroided")[
                self.peaks("centroided")[:, 1] / noise_level
                >= signal_to_noise_threshold
            ]
        if len(self.peaks("raw")) != 0:
            self._peak_dict["raw"] = self.peaks("raw")[
                self.peaks("raw")[:, 1] / noise_level >= signal_to_noise_threshold
            ]
        self._peak_dict["reprofiled"] = None
        return self

    def estimated_noise_level(self, mode="median"):
        """
        Calculates noise threshold for function remove_noise.

        Different modes are available. Default is 'median'

        Keyword Arguments:
            mode (str): define mode for removing noise. Default = "median"
                (other modes: "mean", "mad")

        Returns:
            noise_level (float): estimate noise threshold


        """
        if len(self.peaks("centroided")) == 0:  # or is None?
            return 0

        self.noise_level_estimate = {}
        if mode not in self.noise_level_estimate.keys():
            if mode == "median":
                self.noise_level_estimate["median"] = np.median(
                    self.peaks("centroided")[:, 1]
                )
            elif mode == "mad":
                median = self.estimated_noise_level(mode="median")
                self.noise_level_estimate["mad"] = np.median(
                    abs(self.peaks("centroided")[:, 1] - median)
                )
            elif mode == "mean":
                self.noise_level_estimate["mean"] = sum(
                    self.peaks("centroided")[:, 1]
                ) / float(len(self.peaks("centroided")))
            else:
                print(
                    "Do not understand noise level estimation method call with given mode: {0}".format(
                        mode
                    )
                )
            return_value = self.noise_level_estimate[mode]
        return return_value

    def highest_peaks(self, n):
        """
        Function to retrieve the n-highest centroided peaks of the spectrum.

        Arguments:
            n (int): number of highest peaks to return.

        Returns:
            centroided peaks (list): list mz, i tupls with n-highest

        Example:

        >>> run = pymzml.run.Reader(
        ...     "tests/data/example.mzML.gz",
        ...      MS_precisions =  {
        ...         1 : 5e-6,
        ...         2 : 20e-6
        ...     }
        ... )
        >>> for spectrum in run:
        ...     if spectrum.ms_level == 2:
        ...         if spectrum.ID == 1770:
        ...             for mz,i in spectrum.highest_peaks(5):
        ...                print(mz, i)

        """
        if self._centroided_peaks_sorted_by_i is None:
            self._centroided_peaks_sorted_by_i = self.peaks("centroided")[
                self.peaks("centroided")[:, 1].argsort()
            ]
        return self._centroided_peaks_sorted_by_i[-n:]

    def ppm2abs(self, value, ppm_value, direction=1, factor=1):
        """
        Returns the value plus (or minus, dependent on direction) the
        error (measured precision ) for this value.

        Arguments:
            value (float)   : m/z value
            ppm_value (int)  : ppm value

        Keyword Arguments:
            direction (int) : plus or minus the considered m/z value.
                The argument *direction* should be 1 or -1
            factor (int)    : multiplication factor for the imprecision.
                The argument *factor* should be bigger than 0

        Returns:
            imprecision (float): imprecision for the given value

        """
        result = value + (value * (ppm_value * factor)) * direction
        return result

    def extreme_values(self, key):
        """
        Find extreme values, minimal and maximum m/z and intensity

        Arguments:
            key (str)       : m/z : "mz" or  intensity : "i"

        Returns:
            extrema (tuple) : tuple of minimal and maximum m/z or intensity

        """
        available_extreme_values = ["mz", "i"]
        if key not in available_extreme_values:
            print(
                "Do not understand extreme request: '{0}'; available values are: {1}".format(
                    key, available_extreme_values
                )
            )
            exit()
        if self._extreme_values is None:
            self._extreme_values = {}
        try:
            if key == "mz":
                all_mz_values = self.peaks("raw")[:, 0]
                self._extreme_values["mz"] = (all_mz_values.min(), all_mz_values.max())
                self._extreme_values["mz"] = (
                    self.peaks("raw")[:, 0].min(),
                    self.peaks("raw")[:, 0].max(),
                )
            else:
                all_i_values = self.peaks("raw")[:, 1]
                self._extreme_values["i"] = (
                    self.peaks("raw")[:, 1].min(),
                    self.peaks("raw")[:, 1].max(),
                )
        except ValueError:
            # emtpy spectrum
            self._extreme_values[key] = ()
        return self._extreme_values[key]

    def has_peak(self, mz2find):
        """
        Checks if a Spectrum has a certain peak.
        Requires a m/z value as input and returns a list of peaks if the m/z
        value is found in the spectrum, otherwise ``[]`` is returned.
        Every peak is a tuple of m/z and intensity.

        Note:
            Multiple peaks may be found, depending on the defined precisions

        Arguments:
            mz2find (float): m/z value which should be found

        Returns:
            peaks (list): list of m/z, i tuples

        Example:

        >>> import pymzml
        >>> example_file = 'tests/data/example.mzML'
        >>> run = pymzml.run.Reader(
        ...     example_file,
        ...     MS_precisions =  {
        ...         1 : 5e-6,
        ...         2 : 20e-6
        ...     }
        ... )
        >>> for spectrum in run:
        ...     if spectrum.ms_level == 2:
        ...             peak_to_find = spectrum.has_peak(1016.5404)
        ...             print(peak_to_find)
        [(1016.5404, 19141.735187697403)]

        """
        value = self.transform_mz(mz2find)
        return self.transformed_mz_with_error[value]

    def has_overlapping_peak(self, mz):
        """
        Checks if a spectrum has more than one peak for a given m/z value
        and within the measured precision

        Arguments:
            mz (float): m/z value which should be checked

        Returns:
            Boolean (bool): Returns ``True`` if a nearby peak is detected,
            otherwise ``False``
        """
        for minus_or_plus in [-1, 1]:
            target = self.ppm2abs(mz, self.measured_precision, minus_or_plus, 1)
            temp = self.has_peak(self.ppm2abs(mz, self.measured_precision))
            if temp and len(temp) > 1:
                return True
        return False

    def similarity_to(self, spec2, round_precision=0):
        """
        Compares two spectra and returns cosine

        Arguments:
            spec2 (Spectrum): another pymzml spectrum that is compared to the
                current spectrum.

        Keyword Arguments:
            round_precision (int): precision mzs are rounded to, i.e. round( mz,
                round_precision )

        Returns:
            cosine (float): value between 0 and 1, i.e. the cosine between the
                two spectra.

        Note:
            Spectra data is transformed into an n-dimensional vector,
            where m/z values are binned in bins of 10 m/z and the intensities
            are added up. Then the cosine is calculated between those two
            vectors. The more similar the specs are, the closer the value is
            to 1.
        """
        assert isinstance(spec2, Spectrum), "Spectrum 2 is not a pymzML spectrum"

        vector1 = ddict(int)
        vector2 = ddict(int)
        mzs = set()
        for mz, i in self.peaks("raw"):
            vector1[round(mz, round_precision)] += i
            mzs.add(round(mz, round_precision))
        for mz, i in spec2.peaks("raw"):
            vector2[round(mz, round_precision)] += i
            mzs.add(round(mz, round_precision))

        z = 0
        n_v1 = 0
        n_v2 = 0

        for mz in mzs:
            int1 = vector1[mz]
            int2 = vector2[mz]
            z += int1 * int2
            n_v1 += int1 * int1
            n_v2 += int2 * int2
        try:
            cosine = z / (math.sqrt(n_v1) * math.sqrt(n_v2))
        except:
            cosine = 0.0
        return cosine

    def transform_mz(self, value):
        """
        pymzml uses an internal precision for different tasks. This precision
        depends on the measured precision and is calculated when
        :py:attr:`spec.Spectrum.measured_precision` is invoked. transform_mz
        can be used to transform m/z values into the internal standard.

        Arguments:
            value (float): m/z value

        Returns:
            transformed value (float): to internal standard transformed mz
            value this value can be used to probe internal dictionaries,
            lists or sets, e.g. :func:`pymzml.spec.Spectrum.t_mz_set`

        Example:

            >>> import pymzml
            >>> run = pymzml.run.Reader(
            ...     "test.mzML.gz" ,
            ...     MS_precisions =  {
            ...         1 : 5e-6,
            ...         2 : 20e-6
            ...     }
            ... )
            >>>
            >>> for spectrum in run:
            ...     if spectrum.ms_level == 2:
            ...         peak_to_find = spectrum.has_deconvoluted_peak(
            ...             1044.5804
            ...         )
            ...         print(peak_to_find)
            [(1044.5596, 3809.4356300564586)]

        """
        return int(round(value * self.internal_precision))

    def deprecation_warning(self, function_name):
        deprecation_lookup = {
            "similarityTo": "similarity_to",
            "hasPeak": "has_peak",
            "extremeValues": "extreme_values",
            "transformMZ": "transform_mz",
            "hasOverlappingPeak": "has_overlapping_peak",
            "highestPeaks": "highest_peaks",
            "estimatedNoiseLevel": "estimated_noise_level",
            "removeNoise": "remove_noise",
            "newPlot": "new_plot",
            "centroidedPeaks": "peaks",
            "precursors": "selected_precursors",
        }
        warnings.warn(
            """
            Function: "{0}" deprecated since version 1.0.0, please use new function: "{1}"
            """.format(
                function_name, deprecation_lookup.get(function_name, "not_defined_yet")
            ),
            DeprecationWarning,
        )

    def similarityTo(self, spec2, round_precision=0):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.similarity_to(spec2, round_precision=round_precision)

    def hasPeak(self, mz):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.has_peak(mz)

    def extremeValues(self, key):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.extreme_values(key)

    def transformMZ(self, value):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.transform_mz(value)

    def hasOverlappingPeak(self, mz):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.has_overlapping_peak(mz)

    def highestPeaks(self, n):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.highest_peaks(n)

    def estimatedNoiseLevel(self, mode="median"):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.estimated_noise_level(mode=mode)

    def removeNoise(self, mode="median", noiseLevel=None):
        self.deprecation_warning(sys._getframe().f_code.co_name)
        return self.remove_noise(mode=mode, noise_level=noiseLevel)

    @property
    def centroidedPeaks(self):
        # self.deprecation_warning( sys._getframe().f_code.co_name )
        return self.peaks("centroided")


class Chromatogram(MS_Spectrum):
    """
    Class for Chromatogram access and handling.
    """

    def __init__(self, element, measured_precision=5e-6, *, obo_version=None):
        """
        Arguments:
            element (xml.etree.ElementTree.Element): spectrum as xml Element

        Keyword Arguments:
            measured_precision (float): in ppm, i.e. 5e-6 equals to 5 ppm.
            param (dict): parameter mapping for this spectrum
        """
        self._measured_precision = measured_precision
        self.element = element
        self.noise_level_estimate = {}
        # Property variables
        self._time = None
        self._ms_level = None
        self._i = None
        self._t_mass_set = None
        self._peaks = None
        self._t_mz_set = None
        self._centroided_peaks = None
        self._reprofiled_peaks = None
        self._deconvoluted_peaks = None
        self._profile = None
        self._extreme_values = None
        self._centroided_peaks_sorted_by_i = None
        self._transformed_mz_with_error = None
        self._transformed_mass_with_error = None
        self._precursors = None
        self._ID = None
        self.obo_translator = OboTranslator.from_cache(obo_version)

        if self.element:
            # self._read_accessions()
            self.ns = (
                re.match(r"\{.*\}", element.tag).group(0)
                if re.match(r"\{.*\}", element.tag)
                else ""
            )
            # self._ns_paths            = {
            #     'mz'      : "{ns}binaryDataArrayList/" \
            #                 "{ns}binaryDataArray/" \
            #                 "{ns}cvParam[@accession='{Acc}']/..".format(
            #                     ns=self.ns,
            #                     Acc=self.accessions['time array']
            #             ),
            #     'i'       : "{ns}binaryDataArrayList/" \
            #                 "{ns}binaryDataArray/" \
            #                 "{ns}cvParam[@accession='{Acc}']/..".format(
            #                         ns=self.ns,
            #                         Acc=self.accessions['intensity array']
            #                 ),
            #     'time'    : "{ns}binaryDataArrayList/" \
            #                 "{ns}binaryDataArray/" \
            #                 "{ns}cvParam[@accession='{Acc}']/..".format(
            #                         ns=self.ns,
            #                         Acc=self.accessions['time array']
            #                 ),
            #     'float_type' : "./{ns}cvParam[@accession='{Acc}']"
            # }

        self._decode = self._decode_to_numpy
        # assign function to create numpy array to list???
        self._array = np.array

    def __repr__(self):
        return "<__main__.Chromatogram object with native ID {0} at {1}>".format(
            self.ID, hex(id(self))
        )

    def __str__(self):
        return "<__main__.Chromatogram object with native ID {0} at {1}>".format(
            self.ID, hex(id(self))
        )

    @property
    def ID(self):
        if self._ID is None:
            self._ID = self.element.get("id")
        return self._ID

    @property
    def mz(self):
        """"""
        print("Chromatogram has no property mz.\nReturn retention time instead")
        return self.time

    @property
    def time(self):
        """
        Returns the list of time values. If the time values are encoded, the
        function _decode() is used to decode the encoded data.\n
        The time property can also be set, e.g. for theoretical data.
        However, it is recommended to use the profile property to set time and
        intensity tuples at same time.

        Returns:
            time (list): list of time values from the analyzed chromatogram

        """
        if self._time is None:
            params = self._get_encoding_parameters("time array")
            self._time = self._decode(*params)
        return self._time

    @property
    def i(self):
        if self._i is None:
            params = self._get_encoding_parameters("intensity array")
            self._i = self._decode(*params)
        return self._i

    @property
    def profile(self):
        """
        Returns the list of peaks of the chromatogram as tuples (time, intensity).

        Returns:
            peaks (list): list of time, i tuples

        Example:

        >>> import pymzml
        >>> run = pymzml.run.Reader(
        ...     spectra.mzMl.gz,
        ...     MS_precisions = {
        ...         1 : 5e-6,
        ...         2 : 20e-6
        ...     }
        ... )
        >>> for entry in run:
        ...     if isinstance(entry, pymzml.spec.Chromatogram):
        ...         for time, intensity in entry.peaks:
        ...             print(time, intensity)

        Note:
           The peaks property can also be set, e.g. for theoretical data.
           It requires a list of time/intensity tuples.

        """
        if self._profile is None:
            if self._time is None and self._i is None:
                self._profile = []
                for pos, t in enumerate(self.time):
                    self._profile.append([t, self.i[pos]])
                    # much faster than zip ... list(zip(self.mz, self.i))
            elif self._time is not None and self._i is not None:
                self._profile = []
                for pos, t in enumerate(self.time):
                    self._profile.append([t, self.i[pos]])
            elif self._profile is None:
                self._profile = []
        return self._array(self._profile)

    @profile.setter
    def profile(self, tuple_list):
        if len(tuple_list) == 0:
            return
        # self._mz, self._i = map(list, zip(*tuple_list))
        # same here .. zip is soooooo slow :)
        self._time = []
        self._i = []
        for time, i in tuple_list:
            self._time.append(time)
            self._i.append(i)
        self._peaks = tuple_list
        self._reprofiledPeaks = None
        self._centroidedPeaks = None
        return self

    def peaks(self):
        """
        Return the list of peaks of the spectrum as tuples (time, intensity).

        Returns:
            peaks (list): list of time, intensity tuples

        Example:

        >>> import pymzml
        >>> run = pymzml.run.Reader(
        ...     spectra.mzMl.gz,
        ...     MS_precisions =  {
        ...         1 : 5e-6,
        ...         2 : 20e-6
        ...     }
        ... )
        >>> for entry in run:
        ...     if isinstance(entry, pymzml.spec.Chromatogram):
        ...         for time, intensity in entry.peaks:
        ...             print(time, intensity)

        Note:
           The peaks property can also be set, e.g. for theoretical data.
           It requires a list of time/intensity tuples.

        """
        return self.profile



import os

DATA_FOLDER = os.path.join(
    *[
        os.path.dirname(__file__),
        # 'tests',
        "data",
    ]
)

DATA_FILES = [
    "example.mzML",
    "example.mzML.gz",
    "example.mzML.idx.gz",
    "mini.chrom.mzML",
    "mini.chrom.mzML.gz",
    "mini.chrom.mzML.idx.gz",
    "mini_numpress.chrom.mzML",
    "mini_numpress.chrom.mzML.gz",
    "mini_numpress.chrom.mzML.idx.gz",
    "BSA1.mzML.gz",
    "example_invalid_obo_version.mzML",
    "example_no_obo_version.mzML",
]

paths = [os.path.join(DATA_FOLDER, file) for file in DATA_FILES]



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import argparse
from collections import defaultdict

import pymzml.obo


FIELDNAMES = ["id", "name", "def", "is_a"]


def main(args):
    """
    Use this script to interrogate the OBO database files.

    usage:

        ./queryOBO.py [-h] [-v VERSION] query

    Example::

      $ ./queryOBO.py'scan time'
      MS:1000016
      scan time
      'The time taken for an acquisition by scanning analyzers.' [PSI:MS]
      Is a: MS:1000503 ! scan attribute

    Example::

        $ ./queryOBO.py 1000016
        MS:1000016
        scan time
        "The time taken for an acquisition by scanning analyzers." [PSI:MS]
        MS:1000503 ! scan attribute


    """
    obo = pymzml.obo.OboTranslator(version=args.version)
    obo.parseOBO()
    if args.query.isdigit():
        print(search_by_id(obo, args.query))
    else:
        for ix, match in enumerate(search_by_name(obo, args.query)):
            print("#{0}".format(ix))

            for fieldname in ("id", "name", "def"):
                print(match[fieldname])

            if "is_a" in match:
                print("Is a:", match["is_a"])


def search_by_name(obo, name):
    print("Searching for {0}".format(name.lower()))
    matches = []
    for lookup in obo.lookups:
        for key in lookup.keys():
            if name.lower() in key.lower():
                match = defaultdict(str)

                for fieldname in FIELDNAMES:
                    if fieldname in lookup[key].keys():
                        match[fieldname] = lookup[key][fieldname]

                matches.append(match)

    return matches


def search_by_id(obo, id):
    key = "MS:{0}".format(id)
    return_value = ""
    for lookup in obo.lookups:
        if key in lookup:
            if obo.MS_tag_regex.match(key):
                for fn in FIELDNAMES:
                    if fn in lookup[key].keys():
                        return_value += "{0}\n".format(lookup[key][fn])
    return return_value


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(usage=__doc__)
    argparser.add_argument(
        "query", help="an accession or part of an OBO term name to look for"
    )
    argparser.add_argument(
        "-v",
        "--version",
        default="1.1.0",
        help="""
            the version of the OBO to use; valid options are 1.0.0, 1.1.0, and 1.2,
            default is 1.1.0
        """,
    )

    args = argparser.parse_args()

    main(args)



#!/usr/bin/env python
import sys
import pymzml


def main(mzml_file):
    """
    Basic example script to demonstrate the usage of pymzML. Requires a mzML 
    file as first argument.

    usage:
    
        ./simple_parser.py <path_to_mzml_file>

    Note:

        This script uses the new syntax with the MS level being a property of
        the spectrum class ( Spectrum.ms_level ). The old syntax can be found in
        the script simple_parser_v2.py where the MS level can be queried as a key
        (Spectrum['ms level'])


    """
    run = pymzml.run.Reader(mzml_file)
    for n, spec in enumerate(run):
        print(
            "Spectrum {0}, MS level {ms_level} @ RT {scan_time:1.2f}".format(
                spec.ID, ms_level=spec.ms_level, scan_time=spec.scan_time_in_minutes()
            )
        )
    print("Parsed {0} spectra from file {1}".format(n, mzml_file))
    print()


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(main.__doc__)
        exit()
    mzml_file = sys.argv[1]
    main(mzml_file)



#!/usr/bin/env python

"""

Check for presence of example files and download them.
Just call the function open_example()

For using the correct output directory just state get_example_file.OUTPUT_DIR

"""
import hashlib
import os.path
import sys


SHA256_DICT = {
    "deconvolution.mzML.gz": (
        "19399e4f87d8937cf7be32bab7fd8889700e119cbc2ede215270cbfc9913b394",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
    "dta_example.mzML": (
        "2cd7207d5106ee5fb3ab7867961a6c1ad02f0ef72da19532e7ebb41c8d010b1a",
        "http://dev.thep.lu.se/fp6-prodac/export/80/trunk/mzML/",
    ),
    "neutral_loss_example_1.1.0.mzML": (
        "5911fa5d91a435aedd7be5a2ea09272f437d44f350948179a96146705a177a16",
        "http://dev.thep.lu.se/fp6-prodac/export/80/trunk/mzML/scans/",
    ),
    "precursor_spectrum_example_1.1.0.mzML": (
        "783187ec7eb2ee2713aecd9f6032508caad3439afa0b4a20232480b33472c17b",
        "http://dev.thep.lu.se/fp6-prodac/export/80/trunk/mzML/scans/",
    ),
    "profile-mass-spectrum.mzml": (
        "2e952de2a25db421f3e30d8c659acc6c3f5e862e4d0fd32ec24c7c606c9a62fd",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
    "small-1.0.0.mzml": (
        "9e15d6d2eba26932e4d5c61f92df7771da0bbb8589b7237d05bd915017dbf82f",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
    "BSA1.mzML": (
        "d4bde93c77ec9e948cc62f4c022b8d54591073fd1170e264b69a79dc8d259830",
        "http://open-ms.svn.sourceforge.net/viewvc/open-ms/OpenMS/share/OpenMS/examples/BSA/",
    ),
    "proteomeDiscoverer.mzML.gz": (
        "7fdd45bfcb6d82b2a75abd63fdc0d8cc1f14c1bd717ba2f3d4eec87ec201f3be",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
    "32-bit.mzML.gz": (
        "199d2611c9894f2a07fb734fbe4db2226cd2a43e806d4e9662b66b979f7df264",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
    "32-bit-compressed.mzML.gz": (
        "30fdde9ebf3bb145a60ea589f466cb0db8b3b2c451646156db3e56ffe1bf0e59",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
    "64-bit.mzML.gz": (
        "0d56a2295096e7499a8a90509661d22d4a6a097ad58f3c03310804c9698113aa",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
    "small.pwiz.1.1.mzML": (
        "87ddde776c9933857edb22f404116f23e85ca8bddadc84d10308eb8853a58768",
        "http://proteowizard.sourceforge.net/example_data/",
    ),
    "MRM_example_1.1.0.mzML": (
        "8140d77921be6fe49e73d71536b211ad39ae48f0272d4a465da2dd202185bdd9",
        "http://dev.thep.lu.se/fp6-prodac/export/80/trunk/mzML/scans/",
    ),
    "emptyScan.mzML.gz": (
        "4d82ea02c1f938e8fbe97b907371aad0eceee074bbee37f7513374a1bb7c3f3b",
        "http://www.uni-muenster.de/hippler/pymzml/example_mzML_files/",
    ),
}

EXAMPLE_DIR = os.path.join(os.path.dirname(__file__), "mzML_example_files")
OUTPUT_DIR = os.path.join(os.path.dirname(__file__), "output")

try:
    os.makedirs(EXAMPLE_DIR)
except OSError:
    pass
try:
    os.makedirs(OUTPUT_DIR)
except OSError:
    pass


def open_example(filename):
    """
    filename = filename of a file in the example directory
    If the file "filename" is not present in the example directory, the file is downloaded.
    Path with filename is returned.
    """
    try:
        SHA256_DICT[filename]
    except KeyError:
        print('File "{0}" not available.'.format(filename), file=sys.stderr)
        exit(1)

    filename_with_path = os.path.join(EXAMPLE_DIR, filename)
    # check file presence
    checked_file = check_file(filename_with_path, filename)
    if checked_file:
        return filename_with_path

    # file is not present or corrupt
    server_url = SHA256_DICT[filename][1]
    url = server_url + filename
    if download(filename_with_path, url):
        checked_file = check_file(filename_with_path, filename)
        if checked_file:
            return filename_with_path
        else:
            os.remove(filename_with_path)
            print("Downloaded file is corrupt. Please try again.", file=sys.stderr)
            exit(1)

        exit(1)


def check_file(filename_with_path, filename):
    # check file presence
    if os.path.isfile(filename_with_path):
        # file exists, check SHA224 sum
        file = open(filename_with_path, "rb")
        if check_sha256(file, filename):
            return True


def check_sha256(file, filename):
    sha256_sum = SHA256_DICT[filename][0]

    if hashlib.sha256(file.read()).hexdigest() == sha256_sum:
        return True
    else:
        return False


def calc_sha256(file):
    file = open(file, "rb")
    return hashlib.sha256(file.read()).hexdigest()


def download(filename_with_path, url):
    """
    filename_with_path = path and filename, where to save downloaded file
    url = url to download
    Path with filename is returned.
    """

    if sys.version_info[0] < 3:
        from urllib2 import Request, urlopen, URLError, HTTPError
    else:
        from urllib.request import Request, urlopen, URLError, HTTPError

    print("Downloading file ...", end="\r", file=sys.stderr)
    req = Request(url)
    print("Downloading file completed.", file=sys.stderr)
    try:
        response = urlopen(req)
        with open(filename_with_path, "wb") as f:
            f.write(response.read())
        return filename_with_path
    except HTTPError:
        print(
            "HTTP Error: The server could not fulfill the request.",
            "Possibly, the file is not found on the server.",
            file=sys.stderr,
        )
        exit(1)
    except URLError:
        print(
            "URL Error: Server not available. Please check your internet connection.",
            file=sys.stderr,
        )
        exit(1)


if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python
# -*- coding: latin-1 -*-
"""
Reader class for indexed gzipped files
"""

# Python mzML module - pymzml
# Copyright (C) 2010-2019 M. Kösters, C. Fufezan
#     The MIT License (MIT)

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

import struct
import zlib
from collections import OrderedDict


class GSGR(object):
    """
    Generalized Gzip reader class which enables random access in files
    written with the :class:`~pymzml.utils.GSGW.GSGW` class.

    Keyword Arguments:
        file (str): path to file to read
    """

    def __init__(self, file=None):

        self.file_in = open(file, "rb")
        self.filename = file
        self.magic_bytes = b"\x1f\x8b"
        self.indexed = True

        if not self._check_magic_bytes():
            raise Exception("not a gzip file (wrong magic bytes)")

        self.random_access = False  # initial state, until index is read

        self._read_basic_header()
        if self.flg & 0 != 0:  # FTEXT flag
            self.ascii_file = True
        if self.flg & 2 != 0:  # FHCRC flag
            crc16 = self.file_in.read(2)
        if self.flg & 4 != 0:  # FEXTRA flag
            # TODO: maybe never tested
            xlen = struct.unpack("<H", self.file_in.read(2))[0]
            self.file_in.seek(xlen)
        if self.flg & 8 != 0:  # FNAME flag
            self.fname = self._read_until_zero()
        if self.flg & 16 == 0:  # FCOMMENT flag NOT SET
            self.indexed = False
        else:
            self._read_index()

    def __del__(self):
        try:
            self.close()
        except:
            raise Exception(" cant close file")

    def seek(self, offset):
        """
        Seek to byte offset in input file.

        Arguments:
            offset (int): byte offset to seek to in FileIn

        Returns:
            None
        """
        self.file_in.seek(offset)
        return

    def read_block(self, index):
        """
        Read and return the data block with the unique index `index`

        Arguments:
            index(int or str): identifier associated with a specific block

        Returns:
            data (str): indexed text block as string
        """
        start = self.index[index]
        try:
            end = self.index[int(index) + 1]
        except:
            end = self.file_in.seek(0, 2)
        self.file_in.seek(start)
        readSize = end - start
        comp_data = self.file_in.read(readSize)
        data = zlib.decompress(comp_data, -zlib.MAX_WBITS)
        return data

    def _check_magic_bytes(self):
        """
        Check if file is a gzip file.
        """
        # self.file_in.seek(0) # make sure file pointer is at start
        mb = self.file_in.read(2)
        return mb == self.magic_bytes

    def _read_basic_header(self):
        """
        Read and save compression method, bitflags, changetime,
        compression speed and os.
        """
        self.file_in.seek(2)  # make sure filepoiner is at correct position
        vals = struct.unpack("<BBLBB", self.file_in.read(8))
        self.cm = vals[0]
        self.flg = vals[1]
        self.mtime = vals[2]
        self.xfl = vals[3]
        self.os = vals[4]

    def _read_until_zero(self):
        """
        Read input until \x00 is reached
        """
        buf = b""
        c = self.file_in.read(1)
        while c != b"\x00":
            buf += c
            c = self.file_in.read(1)
        return buf

    def _read_index(self):
        """
        Read and save offset dict from indexed gzip file
        """
        self.index = OrderedDict()
        self.file_in.seek(10)  # make sure file pointer is at right position
        mb = self.file_in.read(3)
        if mb != b"FU\x01":  # All hail MK!
            print("No index in comment field found. No random access possible")
            self.indexed = False
        lengths = struct.unpack("<BB", self.file_in.read(2))
        self.idx_len = lengths[0]
        self.offset_len = lengths[1]
        ID_block = b""
        while b"\x00" not in ID_block:
            ID_block = self.file_in.read(self.idx_len)
            OffsetBlock = self.file_in.read(self.offset_len)
            try:
                try:
                    Identifier = int(ID_block.decode("latin-1").strip("¬"))
                except:
                    Identifier = ID_block.decode("latin-1").strip("¬")
                Offset = int(OffsetBlock.decode("latin-1").strip("¬"))
                self.index[Identifier] = Offset
            except:
                break
        self.file_in.seek(0)

    def read(self, size=-1):
        """
        Read the content of the in File in binary mode

        Keyword Arguments:
            size (int, optional): number of bytes to read, -1 for everything

        Returns:
            data (bytes): parsed bytes from input file
        """
        return self.file_in.read(size)

    def __enter__(self):
        """
        Enable the with syntax for this class (entry point)
        """
        return self.file_in

    def __exit__(self, exc_type, exc_value, traceback):
        """
        destructor when using this class with 'with .. as '
        """
        self.file_in.close()

    def close(self):
        """
        Close the internal Filehandler
        """
        self.file_in.close()


if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Part of pymzml test cases
"""

import sys
import os

sys.path.append(os.path.abspath("."))
import unittest
import gzip


class TestRound(unittest.TestCase):
    def _check_rounding(self, test, expected):
        assert round(test) == expected

    def test_rounding(self):
        for x, y in [(1, 1), (1.9, 2)]:
            yield self._check_rounding, x, y


class TestOboVersion(unittest.TestCase):
    """
    Test that all obo filenames match the version contained

    """

    def setUp(self):
        self.obodir = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), os.pardir, "pymzml", "obo"
        )
        self.obo_files = [f for f in os.listdir(self.obodir) if f.startswith("psi-ms")]

    def _check_version(self, a, b):
        assert a == b

    def _get_file_version(self, file_path):
        v = ""
        with gzip.open(file_path, "r") as f:
            for line in f:
                if line.decode().strip().startswith("data-version"):
                    v = line.decode().split(":")[-1].strip()
                if line.decode().startswith("remark: version"):
                    v = line.decode().split(":")[-1].strip()
                    break
        return v

    def test_version(self):
        for fn in self.obo_files:
            v = self._get_file_version(os.path.join(self.obodir, fn))
            _v = fn[7:-7].strip()
            yield self._check_version, v, _v


if __name__ == "__main__":
    unittest.main(verbosity=3)



#!/usr/bin/env python3
import bisect
import pprint
import click
import os
import hashlib

import random
import pickle
import pymzml

from xml.etree.ElementTree import ParseError


@click.command()
@click.argument('input_files', nargs=-1)
@click.option('--shuffle', '-s', is_flag=True)
@click.option('--suffix')
def main(input_files, shuffle, suffix=''):
    print(suffix)
    output_name = f'failing_spectra{suffix}.txt'
    with open(output_name, 'wt') as fout:
        for file in input_files:
            print(file)
            fails = 0
            fail_info = []
            reader = pymzml.run.Reader(file)
            indices = list(range(1, reader.get_spectrum_count()))
            if shuffle:
                random.seed(a=1570010244.2805517)
                random.shuffle(indices)
                m = hashlib.md5()
                m.update(pickle.dumps(indices))
                md5_digest = m.hexdigest()
                print(f"Shuffled indices md5: {md5_digest}")
                # exit(1)
            number_of_specs = len(indices)
            basic_off_set_dict = reader.info["file_object"].file_handler.offset_dict.copy()
            basic_seek_list = reader.info["file_object"].file_handler.seek_list.copy()
            for pos, i in enumerate(indices):
                # reader.info["file_object"].file_handler.offset_dict = basic_off_set_dict.copy()
                # reader.info["file_object"].file_handler.seek_list = basic_seek_list.copy()
                current_precentage = 100 * float(pos) / float(number_of_specs)
                print(f"[{current_precentage:0>3.1f}%] Access spectrum {i:<10}".format(), end='\r')
                #spec = reader[i]
                try:
                    spec = reader[i]
                    # 100 / 0
                except Exception as e:
                    fails += 1
                    fout.write(f'{os.path.basename(file)}\t{i}\n')
                    print(f"failed on spec {i}")

                    # pprint.pprint(reader.info["file_object"].file_handler.offset_dict)
                    # pprint.pprint(reader.info["file_object"].file_handler.seek_list)
                    spec = reader[i]
                    exit(1)
                # break
            print()
            perc = fails / len(indices) * 100
            print(f'{fails} of {len(indices)} spectra could not be accessed ({perc}%)')
            print('\n\n')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3.4
# encoding: utf-8
"""
Plot a test spectrum for different plotting styles
"""
import pymzml
import os
import sys
import test_file_paths
from pymzml.plot import Factory
import unittest


class PlotTest(unittest.TestCase):
    """ """

    def setUp(self):
        self.paths = test_file_paths.paths
        self.path = self.paths[2]
        self.Run = pymzml.run.Reader(self.path)
        self.spec = self.Run[6]

        self.layout = {
            "font": {"family": "Arial", "size": 20, "color": "#000000"},
            "autosize": False,
            "width": 1700,
            "height": 700,
            "margin": {"l": 100, "r": 80, "t": 100, "b": 60},
            # 'separatethousands' : False,
            # 'autoexpand':True,
            "xaxis1": {
                "type": "linear",
                "color": "#000000",
                "title": "<i>m/z</i>",
                "titlefont": {"family": "Arial", "size": 24, "color": "#000000"},
                "tickmode": "auto",
                "nticks": 0,
                "tickprefix": "",
                "ticksuffix": "",
                "showticklabels": True,
                "tickfont": {"family": "Arial", "size": 20, "color": "rgb(0, 0, 0)"},
                # "tickangle": "auto",
                "tickformat": "",
                "showexponent": "all",
                "exponentformat": "B",
                "ticklen": 5,
                "tickwidth": 1,
                "tickcolor": "#000000",
                "ticks": "outside",
                "showline": True,
                "mirror": False,
                "showgrid": False,
                "zerolinecolor": "rgb(0, 0, 0)",
                "zerolinewidth": 1,
                "zeroline": False,
                "anchor": "y",
                "side": "bottom",
                "tick0": 0,
                "dtick": 100,
            },
            "yaxis1": {
                "type": "linear",
                "color": "#000000",
                "title": "Intensity",
                "titlefont": {"family": "Arial", "size": 24, "color": "#000000"},
                "nticks": 0,
                "tickprefix": "",
                "ticksuffix": "",
                "showticklabels": True,
                "tickfont": {"family": "Arial", "size": 20, "color": "rgb(0, 0, 0)"},
                # "tickangle": "auto",
                "tickformat": "",
                "showexponent": "all",
                "exponentformat": "B",
                # 'separatethousands' : True,
                "ticklen": 5,
                "tickwidth": 1,
                "tickcolor": "#000000",
                "ticks": "outside",
                "showline": True,
                "mirror": False,
                "showgrid": False,
                "zerolinecolor": "rgb(0, 0, 0)",
                "zerolinewidth": 1,
                "zeroline": False,
                "anchor": "x",
                "side": "left",
                "tick0": 0,
                "dtick": 20000000,
            },
            "legend": {
                "bgcolor": "#fff",
                "bordercolor": "#444",
                "borderwidth": 0,
                "font": {"family": "Arial", "size": 20, "color": "#000000"},
                "orientation": "v",
                "traceorder": "normal",
                "x": 1.02,
                "xanchor": "left",
                "y": 1,
                "yanchor": "auto",
            },
        }
        self.test_styles = [
            ("lines", (200, 200, 200), "solid", 0.8, None, None),
            ("sticks", (200, 0, 0), "solid", 0.8, None, (200, 300)),
            ("points", (0, 0, 200), "solid", 0.8, None, None),
            ("label.sticks", (0, 0, 200), "dot", 0.8, None, None),
            ("triangle.big", (200, 0, 200), "solid", 0.8, None, None),
            ("triangle.small", (200, 0, 0), "solid", 0.8, None, None),
            ("label.triangle.MS_precision", (0, 200, 0), "dash", 0.8, None, None),
            ("label.spline", (0, 200, 0), "dash", 0.3, None, None),
            ("label.triangle.small", (0, 200, 0), "solid", 0.8, None, None),
            ("label.linear.bottom", (0, 200, 100), "solid", 0.8, self.layout, None),
        ]
        self.pf = Factory()
        self.file_name = os.path.join(os.path.dirname(self.path), "test_plot.html")

    def test_new_plot(self):
        self.pf.new_plot(MS_precision="100e-6", title="test")
        self.assertEqual(self.pf.MS_precisions[0], "100e-6")
        self.assertEqual(self.pf.titles[0], "test")
        self.pf.new_plot()
        self.assertEqual(len(self.pf.plots), 2)
        self.assertEqual(self.pf.MS_precisions[1], "5e-6")
        self.assertEqual(self.pf.titles[1], None)
        self.pf = Factory()

    def test_add_plots(self):
        for plotnumber, plot in enumerate(self.test_styles):
            self.pf.new_plot()
            style, color, dash, opacity, layout, mz_range = plot
            split_style = style.split(".")
            if split_style[0] == "label":
                self.pf.add(
                    self.spec.highest_peaks(100),
                    color=(200, 200, 200),
                    style="lines",
                    name="datapoints",
                    title="test spec plot",
                    plot_num=plotnumber,
                    mz_range=mz_range,
                )
                label_list = []
                for peak in self.spec.highest_peaks(100):
                    label_list.append((peak[0], 100, "spline_top", "label"))
                self.pf.add(
                    label_list,
                    color=color,
                    style=style,
                    name="Label",
                    opacity=opacity,
                    dash=dash,
                    title="test spec plot",
                    plot_num=plotnumber,
                    mz_range=mz_range,
                )
                self.assertEqual(len(self.pf.plots[plotnumber]), 2)
            else:
                peak_list = self.spec.highest_peaks(100)
                self.pf.add(
                    peak_list,
                    color=color,
                    style=style,
                    name="Label",
                    opacity=opacity,
                    dash=dash,
                    title="test spec plot",
                    plot_num=plotnumber,
                    mz_range=mz_range,
                )

            self.assertEqual(
                self.pf.plots[plotnumber][-1]["line"]["color"],
                "rgba({0},{1},{2},{3})".format(color[0], color[1], color[2], opacity),
            )
            self.assertEqual(self.pf.plots[plotnumber][-1]["line"]["dash"], dash)
            if "label" in style:
                if "spline" in style:
                    mode = "lines+markers+text"
                else:
                    mode = "text+lines"
            elif style == "points":
                mode = "markers"
            else:
                mode = "lines"
            self.assertEqual(self.pf.plots[plotnumber][-1]["mode"], mode)
            if mz_range is not None:
                self.assertLessEqual(self.pf.x_max[plotnumber], mz_range[1])
        self.assertEqual(len(self.pf.plots), len(self.test_styles))

    def test_save_plot(self):
        self.pf.add(
            self.spec.highest_peaks(100),
            color=(200, 200, 200),
            style="sticks",
            name="datapoints",
        )
        self.pf.save(filename=self.file_name, layout=self.layout)
        self.assertTrue(os.path.exists(self.file_name))

    def tearDown(self):
        if os.path.exists(self.file_name):
            os.remove(self.file_name)


if __name__ == "__main__":
    unittest.main(verbosity=3)



#!/usr/bin/env python

import pymzml
from collections import defaultdict as ddict
import os


def main():
    """
    Testscript to isolate the n-highest peaks from an example file.

    Usage:

        ./highest_peaks.py

    Parses the file '../tests/data/example.mzML' and extracts the 2 highest
    intensities from each spectrum.

    """
    example_file = os.path.join(
        os.path.dirname(__file__), os.pardir, "tests", "data", "example.mzML"
    )
    run = pymzml.run.Reader(example_file)
    highest_i_dict = ddict(list)

    number_of_peaks_to_extract = 2

    for spectrum in run:
        # print( spectrum.ID )
        if spectrum.ms_level == 1:
            for mz, i in spectrum.highest_peaks(number_of_peaks_to_extract):
                highest_i_dict[spectrum.ID].append(i)
    for spectrum_id, highest_peak_list in highest_i_dict.items():
        assert len(highest_peak_list) == number_of_peaks_to_extract
        print(
            "Spectrum {0}; highest intensities: {1}".format(
                spectrum_id, highest_peak_list
            )
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os
from operator import itemgetter

import pymzml


def main():
    """
    Extract the 10 most often fragmented precursors from the BSA example file.

    This can e.g. be used for defining exclusion lists for further MS runs.

    usage:

        ./get_precursors.py

    """

    example_file = os.path.join(
        os.path.dirname(__file__), os.pardir, "tests", "data", "BSA1.mzML.gz"
    )
    run = pymzml.run.Reader(example_file)
    fragmented_precursors = {}
    for spectrum in run:
        if spectrum.ms_level == 2:
            selected_precursors = spectrum.selected_precursors
            if spectrum.selected_precursors is not None:
                for precursor_dict in selected_precursors:
                    precursor_mz = precursor_dict["mz"]
                    precursor_i = precursor_dict["i"]
                    rounded_precursor_mz = round(precursor_mz, 3)
                    if rounded_precursor_mz not in fragmented_precursors.keys():
                        fragmented_precursors[rounded_precursor_mz] = []
                    fragmented_precursors[rounded_precursor_mz].append(spectrum.ID)

    precursor_info_list = []
    for rounded_precursor_mz, spectra_list in fragmented_precursors.items():
        precursor_info_list.append(
            (len(spectra_list), rounded_precursor_mz, spectra_list)
        )

    for pos, (number_of_spectra, rounded_precursor_mz, spectra_list) in enumerate(
        sorted(precursor_info_list, reverse=True)
    ):
        print(
            "Found precursor: {0} in spectra: {1}".format(
                rounded_precursor_mz, spectra_list
            )
        )
        if pos > 8:
            break


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import sys
import os
from pymzml.utils.utils import index_gzip
import pymzml
import glob
import multiprocessing


def main(folder, num_cpus=1):
    """
    Creates indexed gzip mzML files from all mzMLs files in the given folder
    using a given number of threads.

    Usage:
        python multi_threading_file_compression.py <folder> <threads>

    Note:
        If the number of threads is larger than the number of actual possible
        threads, all possible threads will be used.

    """
    max_cpus = multiprocessing.cpu_count()
    if int(num_cpus) > max_cpus:
        num_cpus = max_cpus
    else:
        num_cpus = int(num_cpus)
    mzml_job_list = []
    for mzml_path in glob.glob(os.path.join(folder, "*.mzML")):
        out_path = "{0}.gz".format(mzml_path)
        if os.path.exists(out_path):
            print("Skipping: {0}".format(mzml_path))
            continue
        mzml_job_list.append((mzml_path, out_path))
    print(
        "Compressing {0} mzML files using {1} threads".format(
            len(mzml_job_list), num_cpus
        )
    )
    mp_pool = multiprocessing.Pool(num_cpus)
    results = mp_pool.starmap(compress_file, mzml_job_list)
    mp_pool.close()
    print("Done")
    return


def compress_file(file_path, out_path):
    print("Working on file {0}".format(file_path))
    with open(file_path) as fin:
        fin.seek(0, 2)
        max_offset_len = fin.tell()
        max_spec_no = pymzml.run.Reader(file_path).get_spectrum_count() + 10

    index_gzip(
        file_path, out_path, max_idx=max_spec_no, idx_len=len(str(max_offset_len))
    )
    print("Wrote file {0}".format(out_path))
    return


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(main.__doc__)
        exit()
    else:
        main(*sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Part of pymzml test cases
"""
import os
import re
import pymzml.run as run
import unittest
from pymzml.spec import Spectrum, Chromatogram
import test_file_paths


class runTest(unittest.TestCase):
    """ """

    def setUp(self):
        """ """
        self.paths = test_file_paths.paths

        file_compressed_indexed = self.paths[2]
        file_compressed_unindexed = self.paths[1]
        file_uncompressed_indexed = self.paths[0]
        file_uncompressed_unindexed = self.paths[0]
        file_bad_obo_version = self.paths[10]
        file_no_obo_version = self.paths[11]
        self.reader_compressed_indexed = run.Reader(file_compressed_indexed)
        self.reader_compressed_unindexed = run.Reader(file_compressed_unindexed)
        self.reader_uncompressed_indexed = run.Reader(file_uncompressed_indexed)
        self.reader_uncompressed_unindexed = run.Reader(file_uncompressed_unindexed)
        self.reader_bad_obo_version = run.Reader(file_bad_obo_version)
        self.reader_set_obo_version = run.Reader(
            file_bad_obo_version, obo_version="3.25.0"
        )
        self.reader_set_year_obo_version = run.Reader(
            file_uncompressed_indexed, obo_version="23:06:2017"
        )
        self.reader_set_bad_obo_version = run.Reader(
            file_uncompressed_indexed, obo_version="bad_obo_version"
        )
        self.reader_set_no_obo_version = run.Reader(file_no_obo_version)

    def test_with_context(self):
        with run.Reader(self.paths[0]) as reader:
            reader[2]

    def test_setting_measured_precision(self):
        reader = run.Reader(self.paths[2], MS_precisions={1: 61})
        spec = reader[1]
        self.assertEqual(spec.measured_precision, 61)

    def test_determine_file_encoding(self):
        """ """
        encoding = self.reader_compressed_indexed._determine_file_encoding(
            self.paths[2]
        )
        self.assertEqual(encoding, "ISO-8859-1")
        encoding = self.reader_compressed_unindexed._determine_file_encoding(
            self.paths[1]
        )
        self.assertEqual(encoding, "ISO-8859-1")
        encoding = self.reader_uncompressed_indexed._determine_file_encoding(
            self.paths[3]
        )
        self.assertEqual(encoding, "ISO-8859-1")
        encoding = self.reader_uncompressed_unindexed._determine_file_encoding(
            self.paths[0]
        )
        self.assertEqual(encoding, "ISO-8859-1")

    def test_init_iter(self):
        """ """
        mzml_version = self.reader_compressed_indexed.info["mzml_version"]
        obo_version = self.reader_compressed_indexed.info["obo_version"]
        spec_count = self.reader_compressed_indexed.info["spectrum_count"]
        run_id = self.reader_uncompressed_unindexed.info["run_id"]
        start_time = self.reader_uncompressed_unindexed.info["start_time"]
        self.assertEqual(mzml_version, "1.1.0")
        self.assertEqual(obo_version, "3.25.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_compressed_unindexed.info["mzml_version"]
        obo_version = self.reader_compressed_unindexed.info["obo_version"]
        spec_count = self.reader_compressed_unindexed.info["spectrum_count"]
        run_id = self.reader_uncompressed_unindexed.info["run_id"]
        start_time = self.reader_uncompressed_unindexed.info["start_time"]
        self.assertEqual(mzml_version, "1.1.0")
        self.assertEqual(obo_version, "3.25.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_uncompressed_indexed.info["mzml_version"]
        obo_version = self.reader_uncompressed_indexed.info["obo_version"]
        spec_count = self.reader_uncompressed_indexed.info["spectrum_count"]
        run_id = self.reader_uncompressed_unindexed.info["run_id"]
        start_time = self.reader_uncompressed_unindexed.info["start_time"]
        self.assertEqual(mzml_version, "1.1.0")
        self.assertEqual(obo_version, "3.25.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_uncompressed_unindexed.info["mzml_version"]
        obo_version = self.reader_uncompressed_unindexed.info["obo_version"]
        spec_count = self.reader_uncompressed_unindexed.info["spectrum_count"]
        run_id = self.reader_uncompressed_unindexed.info["run_id"]
        start_time = self.reader_uncompressed_unindexed.info["start_time"]

        self.assertEqual(mzml_version, "1.1.0")
        self.assertEqual(obo_version, "3.25.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_bad_obo_version.info["mzml_version"]
        obo_version = self.reader_bad_obo_version.info["obo_version"]
        spec_count = self.reader_bad_obo_version.info["spectrum_count"]
        run_id = self.reader_bad_obo_version.info["run_id"]
        start_time = self.reader_bad_obo_version.info["start_time"]

        self.assertEqual(mzml_version, "1.1.0")
        # run._obo_version_validator 2017 default obo = 4.1.0
        self.assertEqual(obo_version, "4.1.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_set_obo_version.info["mzml_version"]
        obo_version = self.reader_set_obo_version.info["obo_version"]
        spec_count = self.reader_set_obo_version.info["spectrum_count"]
        run_id = self.reader_set_obo_version.info["run_id"]
        start_time = self.reader_set_obo_version.info["start_time"]

        self.assertEqual(mzml_version, "1.1.0")
        self.assertEqual(obo_version, "3.25.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_set_year_obo_version.info["mzml_version"]
        obo_version = self.reader_set_year_obo_version.info["obo_version"]
        spec_count = self.reader_set_year_obo_version.info["spectrum_count"]
        run_id = self.reader_set_year_obo_version.info["run_id"]
        start_time = self.reader_set_year_obo_version.info["start_time"]

        self.assertEqual(mzml_version, "1.1.0")
        # run._obo_version_validator 2017 default obo = 4.1.0
        self.assertEqual(obo_version, "4.1.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_set_bad_obo_version.info["mzml_version"]
        obo_version = self.reader_set_bad_obo_version.info["obo_version"]
        spec_count = self.reader_set_bad_obo_version.info["spectrum_count"]
        run_id = self.reader_set_bad_obo_version.info["run_id"]
        start_time = self.reader_set_bad_obo_version.info["start_time"]

        self.assertEqual(mzml_version, "1.1.0")
        # run._obo_version_validator set invalid obo = 1.1.0
        self.assertEqual(obo_version, "1.1.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

        mzml_version = self.reader_set_no_obo_version.info["mzml_version"]
        obo_version = self.reader_set_no_obo_version.info["obo_version"]
        spec_count = self.reader_set_no_obo_version.info["spectrum_count"]
        run_id = self.reader_set_no_obo_version.info["run_id"]
        start_time = self.reader_set_no_obo_version.info["start_time"]

        self.assertEqual(mzml_version, "1.1.0")
        # run._obo_version_validator set invalid obo = 1.1.0
        self.assertEqual(obo_version, "1.1.0")
        self.assertIsInstance(spec_count, int)
        self.assertEqual(run_id, "exp105-01-ds5562-Pos")
        self.assertEqual(start_time, "2013-09-10T10:31:08Z")

    def test_next(self):
        """ """
        ret = self.reader_compressed_indexed.next()
        self.assertIsInstance(ret, Spectrum)
        ret = self.reader_compressed_unindexed.next()
        self.assertIsInstance(ret, Spectrum)
        ret = self.reader_uncompressed_indexed.next()
        self.assertIsInstance(ret, Spectrum)
        ret = self.reader_uncompressed_unindexed.next()
        self.assertIsInstance(ret, Spectrum)

    def test_get_spec_count(self):
        self.assertEqual(self.reader_compressed_indexed.get_spectrum_count(), 2918)
        self.assertEqual(self.reader_compressed_unindexed.get_spectrum_count(), 2918)
        self.assertEqual(self.reader_uncompressed_unindexed.get_spectrum_count(), 2918)
        self.assertEqual(self.reader_uncompressed_unindexed.get_spectrum_count(), 2918)

    def test_chrom_count_chrom_file(self):
        reader = run.Reader(self.paths[3])
        self.assertEqual(reader.get_chromatogram_count(), 3)

    def test_chrom_count_spec_file(self):
        reader = run.Reader(self.paths[0])
        self.assertEqual(reader.get_chromatogram_count(), None)

    def test_readers_remeber_spawned_spectra(self):
        """
        Make multiple Readers, spawn 10 spectra each, mix them
        and map them back to the reader who spawned them.
        """
        pass

    def test_read_custom_regex(self):
        custom_regex_file = os.path.join(
            os.path.dirname(__file__), "data", "Manuels_custom_ids.mzML"
        )
        reader = run.Reader(
            custom_regex_file,
            index_regex=re.compile(
                b'.*idRef="ManuelsCustomID=(?P<ID>.*) diesdas">(?P<offset>[0-9]*)</offset>'
            ),
        )
        ids = sorted([k for k in reader.info["offset_dict"].keys() if k != "TIC"])
        assert ids == list(range(1, 11))


if __name__ == "__main__":
    unittest.main(verbosity=3)
